# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import pytest

import awkward as ak


def test_NumpyForm():
    form = ak.forms.NumpyForm("float64")

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * float64"
    assert result.tolist() == [0.0]


def test_NumpyForm_inner_shape():
    form = ak.forms.NumpyForm("float64", inner_shape=(2, 3))

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * 2 * 3 * float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * 2 * 3 * float64"
    assert result.tolist() == [[[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]]


def test_EmptyForm():
    form = ak.forms.EmptyForm()

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * unknown"
    assert result.tolist() == []

    with pytest.raises(TypeError, match="cannot generate a length_one_array"):
        form.length_one_array()


def test_RegularForm():
    form = ak.forms.RegularForm(ak.forms.NumpyForm("float64"), 3)

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * 3 * float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * 3 * float64"
    assert result.tolist() == [[0.0, 0.0, 0.0]]


def test_ListOffsetForm():
    form = ak.forms.ListOffsetForm("i64", ak.forms.NumpyForm("float64"))

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * var * float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * var * float64"
    assert result.tolist() == [[]]


def test_ListForm():
    form = ak.forms.ListForm("i64", "i64", ak.forms.NumpyForm("float64"))

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * var * float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * var * float64"
    assert result.tolist() == [[]]


def test_RecordForm():
    form = ak.forms.RecordForm(
        [ak.forms.NumpyForm("float64"), ak.forms.NumpyForm("int64")],
        ["x", "y"],
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * {x: float64, y: int64}"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * {x: float64, y: int64}"
    assert result.tolist() == [{"x": 0.0, "y": 0}]


def test_IndexedForm():
    form = ak.forms.IndexedForm("i64", ak.forms.NumpyForm("float64"))

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * float64"
    assert result.tolist() == [0.0]


def test_IndexedOptionForm():
    form = ak.forms.IndexedOptionForm("i64", ak.forms.NumpyForm("float64"))

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * ?float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * ?float64"
    assert result.tolist() == [None]


def test_ByteMaskedForm():
    form = ak.forms.ByteMaskedForm("i8", ak.forms.NumpyForm("float64"), valid_when=True)

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * ?float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * ?float64"
    assert result.tolist() == [None]


def test_BitMaskedForm():
    form = ak.forms.BitMaskedForm(
        "u8", ak.forms.NumpyForm("float64"), valid_when=True, lsb_order=True
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * ?float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * ?float64"
    assert result.tolist() == [None]


def test_UnmaskedForm():
    form = ak.forms.UnmaskedForm(ak.forms.NumpyForm("float64"))

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * ?float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * ?float64"
    assert result.tolist() == [0.0]


def test_UnionForm():
    form = ak.forms.UnionForm(
        "i8",
        "i64",
        [ak.forms.NumpyForm("float64"), ak.forms.NumpyForm("int64")],
    )

    result = ak.Array(form.length_zero_array())
    assert len(result) == 0

    result = ak.Array(form.length_one_array())
    assert len(result) == 1
    assert result.tolist() == [0.0]


def test_RegularForm_IndexedOptionForm():
    form = ak.forms.RegularForm(
        ak.forms.IndexedOptionForm("i64", ak.forms.NumpyForm("float64")), 10
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * 10 * ?float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * 10 * ?float64"
    assert result.tolist() == [[None] * 10]


def test_RegularForm_ByteMaskedForm():
    form = ak.forms.RegularForm(
        ak.forms.ByteMaskedForm("i8", ak.forms.NumpyForm("float64"), valid_when=True), 5
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * 5 * ?float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * 5 * ?float64"
    assert result.tolist() == [[None] * 5]


def test_RegularForm_BitMaskedForm():
    form = ak.forms.RegularForm(
        ak.forms.BitMaskedForm(
            "u8", ak.forms.NumpyForm("float64"), valid_when=True, lsb_order=True
        ),
        8,
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * 8 * ?float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * 8 * ?float64"
    assert result.tolist() == [[None] * 8]


def test_RegularForm_IndexedForm():
    form = ak.forms.RegularForm(
        ak.forms.IndexedForm("i64", ak.forms.NumpyForm("float64")), 7
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * 7 * float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * 7 * float64"
    assert result.tolist() == [[0.0] * 7]


def test_RegularForm_ListOffsetForm():
    form = ak.forms.RegularForm(
        ak.forms.ListOffsetForm("i64", ak.forms.NumpyForm("float64")), 4
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * 4 * var * float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * 4 * var * float64"
    assert result.tolist() == [[[], [], [], []]]


def test_RegularForm_UnionForm():
    form = ak.forms.RegularForm(
        ak.forms.UnionForm(
            "i8",
            "i64",
            [ak.forms.NumpyForm("float64"), ak.forms.NumpyForm("int64")],
        ),
        3,
    )

    result = ak.Array(form.length_zero_array())
    assert len(result) == 0

    result = ak.Array(form.length_one_array())
    assert len(result) == 1
    assert result.tolist() == [[0.0, 0.0, 0.0]]


def test_RegularForm_RecordForm():
    form = ak.forms.RegularForm(
        ak.forms.RecordForm(
            [ak.forms.NumpyForm("float64"), ak.forms.NumpyForm("int64")],
            ["x", "y"],
        ),
        2,
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * 2 * {x: float64, y: int64}"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * 2 * {x: float64, y: int64}"
    assert result.tolist() == [[{"x": 0.0, "y": 0}, {"x": 0.0, "y": 0}]]


def test_RegularForm_RegularForm_IndexedOptionForm():
    form = ak.forms.RegularForm(
        ak.forms.RegularForm(
            ak.forms.IndexedOptionForm("i64", ak.forms.NumpyForm("float64")), 5
        ),
        3,
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * 3 * 5 * ?float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * 3 * 5 * ?float64"
    assert result.tolist() == [[[None] * 5] * 3]


def test_IndexedOptionForm_EmptyForm():
    form = ak.forms.IndexedOptionForm("i64", ak.forms.EmptyForm())

    result = ak.Array(form.length_zero_array())
    assert len(result) == 0

    result = ak.Array(form.length_one_array())
    assert len(result) == 1
    assert result.tolist() == [None]


def test_RegularForm_IndexedOptionForm_EmptyForm():
    form = ak.forms.RegularForm(
        ak.forms.IndexedOptionForm("i64", ak.forms.EmptyForm()), 5
    )

    result = ak.Array(form.length_zero_array())
    assert len(result) == 0

    result = ak.Array(form.length_one_array())
    assert len(result) == 1
    assert result.tolist() == [[None] * 5]


def test_ByteMaskedForm_EmptyForm():
    form = ak.forms.ByteMaskedForm("i8", ak.forms.EmptyForm(), valid_when=True)

    result = ak.Array(form.length_zero_array())
    assert len(result) == 0

    # ByteMaskedForm cannot hide EmptyForm - content length must equal mask length
    with pytest.raises(TypeError, match="cannot generate a length_one_array"):
        form.length_one_array()


def test_BitMaskedForm_EmptyForm():
    form = ak.forms.BitMaskedForm(
        "u8", ak.forms.EmptyForm(), valid_when=True, lsb_order=True
    )

    result = ak.Array(form.length_zero_array())
    assert len(result) == 0

    # BitMaskedForm cannot hide EmptyForm - content length must equal mask length
    with pytest.raises(TypeError, match="cannot generate a length_one_array"):
        form.length_one_array()


def test_RegularForm_ListForm():
    form = ak.forms.RegularForm(
        ak.forms.ListForm("i64", "i64", ak.forms.NumpyForm("float64")), 3
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * 3 * var * float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * 3 * var * float64"
    assert result.tolist() == [[[], [], []]]


def test_RecordForm_tuple():
    form = ak.forms.RecordForm(
        [ak.forms.NumpyForm("float64"), ak.forms.NumpyForm("int64")],
        None,
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * (float64, int64)"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * (float64, int64)"
    assert result.tolist() == [(0.0, 0)]


def test_RegularForm_BitMaskedForm_large():
    # Test bit packing across byte boundaries (size > 8)
    form = ak.forms.RegularForm(
        ak.forms.BitMaskedForm(
            "u8", ak.forms.NumpyForm("float64"), valid_when=True, lsb_order=True
        ),
        10,
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * 10 * ?float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * 10 * ?float64"
    assert result.tolist() == [[None] * 10]


def test_RegularForm_UnmaskedForm():
    form = ak.forms.RegularForm(ak.forms.UnmaskedForm(ak.forms.NumpyForm("float64")), 4)

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * 4 * ?float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * 4 * ?float64"
    assert result.tolist() == [[0.0, 0.0, 0.0, 0.0]]


def test_ByteMaskedForm_valid_when_false():
    form = ak.forms.ByteMaskedForm(
        "i8", ak.forms.NumpyForm("float64"), valid_when=False
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * ?float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * ?float64"
    assert result.tolist() == [None]


def test_BitMaskedForm_valid_when_false():
    form = ak.forms.BitMaskedForm(
        "u8", ak.forms.NumpyForm("float64"), valid_when=False, lsb_order=True
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * ?float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * ?float64"
    assert result.tolist() == [None]


def test_BitMaskedForm_ListOffsetForm():
    form = ak.forms.BitMaskedForm(
        "u8",
        ak.forms.ListOffsetForm("i64", ak.forms.NumpyForm("float64")),
        valid_when=True,
        lsb_order=True,
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * option[var * float64]"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * option[var * float64]"
    assert result.tolist() == [None]


def test_ByteMaskedForm_ListOffsetForm():
    form = ak.forms.ByteMaskedForm(
        "i8",
        ak.forms.ListOffsetForm("i64", ak.forms.NumpyForm("float64")),
        valid_when=True,
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * option[var * float64]"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * option[var * float64]"
    assert result.tolist() == [None]


def test_BitMaskedForm_ListForm():
    form = ak.forms.BitMaskedForm(
        "u8",
        ak.forms.ListForm("i64", "i64", ak.forms.NumpyForm("float64")),
        valid_when=True,
        lsb_order=True,
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * option[var * float64]"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * option[var * float64]"
    assert result.tolist() == [None]


def test_ByteMaskedForm_ListForm():
    form = ak.forms.ByteMaskedForm(
        "i8",
        ak.forms.ListForm("i64", "i64", ak.forms.NumpyForm("float64")),
        valid_when=True,
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * option[var * float64]"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * option[var * float64]"
    assert result.tolist() == [None]


def test_IndexedOptionForm_ListOffsetForm():
    form = ak.forms.IndexedOptionForm(
        "i64", ak.forms.ListOffsetForm("i64", ak.forms.NumpyForm("float64"))
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * option[var * float64]"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * option[var * float64]"
    assert result.tolist() == [None]


def test_IndexedOptionForm_ListForm():
    form = ak.forms.IndexedOptionForm(
        "i64", ak.forms.ListForm("i64", "i64", ak.forms.NumpyForm("float64"))
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * option[var * float64]"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * option[var * float64]"
    assert result.tolist() == [None]


def test_ListOffsetForm_BitMaskedForm():
    form = ak.forms.ListOffsetForm(
        "i64",
        ak.forms.BitMaskedForm(
            "u8", ak.forms.NumpyForm("float64"), valid_when=True, lsb_order=True
        ),
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * var * ?float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * var * ?float64"
    assert result.tolist() == [[]]


def test_ListOffsetForm_ByteMaskedForm():
    form = ak.forms.ListOffsetForm(
        "i64",
        ak.forms.ByteMaskedForm("i8", ak.forms.NumpyForm("float64"), valid_when=True),
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * var * ?float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * var * ?float64"
    assert result.tolist() == [[]]


def test_ListOffsetForm_IndexedOptionForm():
    form = ak.forms.ListOffsetForm(
        "i64",
        ak.forms.IndexedOptionForm("i64", ak.forms.NumpyForm("float64")),
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * var * ?float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * var * ?float64"
    assert result.tolist() == [[]]


def test_ListForm_BitMaskedForm():
    form = ak.forms.ListForm(
        "i64",
        "i64",
        ak.forms.BitMaskedForm(
            "u8", ak.forms.NumpyForm("float64"), valid_when=True, lsb_order=True
        ),
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * var * ?float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * var * ?float64"
    assert result.tolist() == [[]]


def test_ListForm_ByteMaskedForm():
    form = ak.forms.ListForm(
        "i64",
        "i64",
        ak.forms.ByteMaskedForm("i8", ak.forms.NumpyForm("float64"), valid_when=True),
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * var * ?float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * var * ?float64"
    assert result.tolist() == [[]]


def test_ListForm_IndexedOptionForm():
    form = ak.forms.ListForm(
        "i64",
        "i64",
        ak.forms.IndexedOptionForm("i64", ak.forms.NumpyForm("float64")),
    )

    result = ak.Array(form.length_zero_array())
    assert str(result.type) == "0 * var * ?float64"
    assert result.tolist() == []

    result = ak.Array(form.length_one_array())
    assert str(result.type) == "1 * var * ?float64"
    assert result.tolist() == [[]]



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

import numpy as np
import pytest
from packaging.version import parse as parse_version

import awkward as ak


@pytest.mark.skipif(
    parse_version(np.__version__) < parse_version("2.0.0"),
    reason="NumPy 2 is required for the copy kwarg in np.asarray",
)
@pytest.mark.parametrize(
    "iterable",
    (
        [1.1, 2.2, 3.3, 4.4, 5.5],
        [[1.1, 2.2], [3.3, 4.4], [5.5, 6.6]],
        ["one", "two", "three"],
        [b"one", b"two", b"three"],
    ),
)
def test_numpy2(iterable):
    nparray = np.asarray(iterable)
    akarray = ak.highlevel.Array(nparray, check_valid=True)
    out = np.asarray(akarray)
    assert np.array_equal(out, nparray)

    if np.issubdtype(nparray.dtype, np.floating):
        # copy=None and dtype=None, no copy is required
        out = np.asarray(akarray, dtype=None, copy=None)
        assert np.array_equal(out, nparray)
        assert np.shares_memory(akarray.layout.data, nparray)
        assert np.shares_memory(out, nparray)

        # copy=None and the same dtype is requested, no copy is required
        out = np.asarray(akarray, dtype=np.float64, copy=None)
        assert np.array_equal(out, nparray)
        assert np.shares_memory(akarray.layout.data, nparray)
        assert np.shares_memory(out, nparray)

        # changing dtype and copy=None, copy is required
        out = np.asarray(akarray, dtype=np.float32, copy=None)
        assert np.array_equal(out, nparray.astype(np.float32))
        assert np.shares_memory(akarray.layout.data, nparray)
        assert not np.shares_memory(out, nparray)

        # copy=True, copy is enforced
        out = np.asarray(akarray, dtype=None, copy=True)
        assert np.array_equal(out, nparray)
        assert np.shares_memory(akarray.layout.data, nparray)
        assert not np.shares_memory(out, nparray)

        # copy=True and the same dtype is requested, copy is enforced
        out = np.asarray(akarray, dtype=np.float64, copy=True)
        assert np.array_equal(out, nparray)
        assert np.shares_memory(akarray.layout.data, nparray)
        assert not np.shares_memory(out, nparray)

        # copy=True and different dtype is requested, copy is enforced
        out = np.asarray(akarray, dtype=np.float32, copy=True)
        assert np.array_equal(out, nparray.astype(np.float32))
        assert np.shares_memory(akarray.layout.data, nparray)
        assert not np.shares_memory(out, nparray)

        # copy=False and dtype=None, no copy is required
        out = np.asarray(akarray, dtype=None, copy=False)
        assert np.array_equal(out, nparray)
        assert np.shares_memory(akarray.layout.data, nparray)
        assert np.shares_memory(out, nparray)

        # copy=False and the same dtype is requested, no copy is required
        out = np.asarray(akarray, dtype=np.float64, copy=False)
        assert np.array_equal(out, nparray)
        assert np.shares_memory(akarray.layout.data, nparray)
        assert np.shares_memory(out, nparray)

        # copy=False and a different dtype is requested, copy is required
        with pytest.raises(ValueError):
            np.asarray(akarray, dtype=np.float32, copy=False)

        # copy=None and dtype=None, no copy is required
        out = np.array(akarray, dtype=None, copy=None)
        assert np.array_equal(out, nparray)
        assert np.shares_memory(akarray.layout.data, nparray)
        assert np.shares_memory(out, nparray)

        # copy=None and the same dtype is requested, no copy is required
        out = np.array(akarray, dtype=np.float64, copy=None)
        assert np.array_equal(out, nparray)
        assert np.shares_memory(akarray.layout.data, nparray)
        assert np.shares_memory(out, nparray)

        # changing dtype and copy=None, copy is required
        out = np.array(akarray, dtype=np.float32, copy=None)
        assert np.array_equal(out, nparray.astype(np.float32))
        assert np.shares_memory(akarray.layout.data, nparray)
        assert not np.shares_memory(out, nparray)

        # copy=True, copy is enforced
        out = np.array(akarray, dtype=None, copy=True)
        assert np.array_equal(out, nparray)
        assert np.shares_memory(akarray.layout.data, nparray)
        assert not np.shares_memory(out, nparray)

        # copy=True and the same dtype is requested, copy is enforced
        out = np.array(akarray, dtype=np.float64, copy=True)
        assert np.array_equal(out, nparray)
        assert np.shares_memory(akarray.layout.data, nparray)
        assert not np.shares_memory(out, nparray)

        # copy=True and different dtype is requested, copy is enforced
        out = np.array(akarray, dtype=np.float32, copy=True)
        assert np.array_equal(out, nparray.astype(np.float32))
        assert np.shares_memory(akarray.layout.data, nparray)
        assert not np.shares_memory(out, nparray)

        # copy=False and dtype=None, no copy is required
        out = np.array(akarray, dtype=None, copy=False)
        assert np.array_equal(out, nparray)
        assert np.shares_memory(akarray.layout.data, nparray)
        assert np.shares_memory(out, nparray)

        # copy=False and the same dtype is requested, no copy is required
        out = np.array(akarray, dtype=np.float64, copy=False)
        assert np.array_equal(out, nparray)
        assert np.shares_memory(akarray.layout.data, nparray)
        assert np.shares_memory(out, nparray)

        # copy=False and a different dtype is requested, copy is required
        with pytest.raises(ValueError):
            np.array(akarray, dtype=np.float32, copy=False)


@pytest.mark.skipif(
    parse_version(np.__version__) >= parse_version("2.0.0"),
    reason="NumPy 1 does not support the copy kwarg in np.asarray but does in np.array",
)
@pytest.mark.parametrize(
    "iterable",
    (
        [1.1, 2.2, 3.3, 4.4, 5.5],
        [[1.1, 2.2], [3.3, 4.4], [5.5, 6.6]],
        ["one", "two", "three"],
        [b"one", b"two", b"three"],
    ),
)
def test_numpy1(iterable):
    nparray = np.asarray(iterable)
    akarray = ak.highlevel.Array(nparray, check_valid=True)
    out = np.asarray(akarray)
    assert np.array_equal(out, nparray)

    if np.issubdtype(nparray.dtype, np.floating):
        # asarray with dtype=None, no copy is required
        out = np.asarray(akarray, dtype=None)
        assert np.array_equal(out, nparray)
        assert np.shares_memory(akarray.layout.data, nparray)
        assert np.shares_memory(out, nparray)

        # asarray with same dtype, no copy is required
        out = np.asarray(akarray, dtype=np.float64)
        assert np.array_equal(out, nparray)
        assert np.shares_memory(akarray.layout.data, nparray)
        assert np.shares_memory(out, nparray)

        # asarray with different dtype, copy is required
        out = np.asarray(akarray, dtype=np.float32)
        assert np.array_equal(out, nparray.astype(np.float32))
        assert np.shares_memory(akarray.layout.data, nparray)
        assert not np.shares_memory(out, nparray)

        # copy=True and dtype=None, copy is enforced
        out = np.array(akarray, dtype=None, copy=True)
        assert np.array_equal(out, nparray)
        assert np.shares_memory(akarray.layout.data, nparray)
        assert not np.shares_memory(out, nparray)

        # copy=True and the same dtype is requested, copy is enforced
        out = np.array(akarray, dtype=np.float64, copy=True)
        assert np.array_equal(out, nparray)
        assert np.shares_memory(akarray.layout.data, nparray)
        assert not np.shares_memory(out, nparray)

        # copy=True and different dtype is requested, copy is enforced
        out = np.array(akarray, dtype=np.float32, copy=True)
        assert np.array_equal(out, nparray.astype(np.float32))
        assert np.shares_memory(akarray.layout.data, nparray)
        assert not np.shares_memory(out, nparray)

        # copy=False and dtype=None, no copy is required
        out = np.array(akarray, dtype=None, copy=False)
        assert np.array_equal(out, nparray)
        assert np.shares_memory(akarray.layout.data, nparray)
        assert np.shares_memory(out, nparray)

        # copy=False and the same dtype is requested, no copy is required
        out = np.array(akarray, dtype=np.float64, copy=False)
        assert np.array_equal(out, nparray)
        assert np.shares_memory(akarray.layout.data, nparray)
        assert np.shares_memory(out, nparray)

        # copy=False and a different dtype is requested, copy is required
        out = np.array(akarray, dtype=np.float32, copy=False)
        assert np.array_equal(out, nparray.astype(np.float32))
        assert np.shares_memory(akarray.layout.data, nparray)
        assert not np.shares_memory(out, nparray)



from __future__ import annotations

import cupy as cp
import numpy as np
import pytest

import awkward as ak
from awkward._behavior import behavior_of
from awkward._nplikes.typetracer import TypeTracer

to_list = ak.operations.to_list


@pytest.fixture(scope="function", autouse=True)
def cleanup_cuda():
    yield
    try:
        cp.cuda.Device().synchronize()  # wait for all kernels
    except cp.cuda.runtime.CUDARuntimeError as e:
        print("GPU error during sync:", e)
    cp._default_memory_pool.free_all_blocks()


def test_0582_propagate_context_in_broadcast_and_apply_firsts():
    array = ak.Array([[[0, 1, 2], []], [[3, 4]], [], [[5], [6, 7, 8, 9]]])
    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(ak.operations.firsts(cuda_array, axis=0)) == [[0, 1, 2], []]
    assert to_list(ak.operations.firsts(cuda_array, axis=1)) == [
        [0, 1, 2],
        [3, 4],
        None,
        [5],
    ]
    assert to_list(ak.operations.firsts(cuda_array, axis=2)) == [
        [0, None],
        [3],
        [],
        [5, 6],
    ]
    assert to_list(ak.operations.firsts(cuda_array, axis=-1)) == [
        [0, None],
        [3],
        [],
        [5, 6],
    ]
    assert to_list(ak.operations.firsts(cuda_array, axis=-2)) == [
        [0, 1, 2],
        [3, 4],
        None,
        [5],
    ]
    assert to_list(ak.operations.firsts(cuda_array, axis=-3)) == [
        [0, 1, 2],
        [],
    ]

    with pytest.raises(ValueError):
        ak.operations.firsts(cuda_array, axis=-4)


def test_0582_propagate_context_in_broadcast_and_apply_toregular():
    array = ak.Array(
        [
            {
                "x": np.arange(2 * 3 * 5).reshape(2, 3, 5).tolist(),
                "y": np.arange(2 * 3 * 5 * 7).reshape(2, 3, 5, 7),
            }
        ]
    )
    cuda_array = ak.to_backend(array, "cuda")

    assert str(cuda_array.type) in (
        "1 * {x: var * var * var * int64, y: var * var * var * var * int64}",
        "1 * {y: var * var * var * var * int64, x: var * var * var * int64}",
    )
    assert str(ak.operations.to_regular(cuda_array, axis=-1).type) in (
        "1 * {x: var * var * 5 * int64, y: var * var * var * 7 * int64}",
        "1 * {y: var * var * var * 7 * int64, x: var * var * 5 * int64}",
    )
    assert str(ak.operations.to_regular(cuda_array, axis=-2).type) in (
        "1 * {x: var * 3 * var * int64, y: var * var * 5 * var * int64}",
        "1 * {y: var * var * 5 * var * int64, x: var * 3 * var * int64}",
    )
    assert str(ak.operations.to_regular(cuda_array, axis=-3).type) in (
        "1 * {x: 2 * var * var * int64, y: var * 3 * var * var * int64}",
        "1 * {y: var * 3 * var * var * int64, x: 2 * var * var * int64}",
    )


def test_0582_propagate_context_in_broadcast_and_apply_cartesian():
    one = ak.Array(np.arange(2 * 3 * 5 * 7).reshape(2, 3, 5, 7).tolist())
    two = ak.Array(np.arange(2 * 3 * 5 * 7).reshape(2, 3, 5, 7).tolist())

    cuda_one = ak.to_backend(one, "cuda")
    cuda_two = ak.to_backend(two, "cuda")

    assert (
        str(ak.operations.cartesian([cuda_one, cuda_two], axis=0, nested=True).type)
        == "2 * 2 * (var * var * var * int64, var * var * var * int64)"
    )
    assert (
        str(ak.operations.cartesian([cuda_one, cuda_two], axis=1, nested=True).type)
        == "2 * var * var * (var * var * int64, var * var * int64)"
    )
    assert (
        str(ak.operations.cartesian([cuda_one, cuda_two], axis=2, nested=True).type)
        == "2 * var * var * var * (var * int64, var * int64)"
    )
    assert (
        str(ak.operations.cartesian([cuda_one, cuda_two], axis=3, nested=True).type)
        == "2 * var * var * var * var * (int64, int64)"
    )
    assert (
        str(ak.operations.cartesian([cuda_one, cuda_two], axis=-1, nested=True).type)
        == "2 * var * var * var * var * (int64, int64)"
    )
    assert (
        str(ak.operations.cartesian([cuda_one, cuda_two], axis=-2, nested=True).type)
        == "2 * var * var * var * (var * int64, var * int64)"
    )
    assert (
        str(ak.operations.cartesian([cuda_one, cuda_two], axis=-3, nested=True).type)
        == "2 * var * var * (var * var * int64, var * var * int64)"
    )
    assert (
        str(ak.operations.cartesian([cuda_one, cuda_two], axis=-4, nested=True).type)
        == "2 * 2 * (var * var * var * int64, var * var * var * int64)"
    )

    with pytest.raises(ValueError):
        ak.operations.cartesian([cuda_one, cuda_two], axis=-5, nested=True)

    assert (
        str(ak.operations.cartesian([cuda_one, cuda_two], axis=0).type)
        == "4 * (var * var * var * int64, var * var * var * int64)"
    )
    assert (
        str(ak.operations.cartesian([cuda_one, cuda_two], axis=1).type)
        == "2 * var * (var * var * int64, var * var * int64)"
    )
    assert (
        str(ak.operations.cartesian([cuda_one, cuda_two], axis=2).type)
        == "2 * var * var * (var * int64, var * int64)"
    )
    assert (
        str(ak.operations.cartesian([cuda_one, cuda_two], axis=3).type)
        == "2 * var * var * var * (int64, int64)"
    )
    assert (
        str(ak.operations.cartesian([cuda_one, cuda_two], axis=-1).type)
        == "2 * var * var * var * (int64, int64)"
    )
    assert (
        str(ak.operations.cartesian([cuda_one, cuda_two], axis=-2).type)
        == "2 * var * var * (var * int64, var * int64)"
    )
    assert (
        str(ak.operations.cartesian([cuda_one, cuda_two], axis=-3).type)
        == "2 * var * (var * var * int64, var * var * int64)"
    )
    assert (
        str(ak.operations.cartesian([cuda_one, cuda_two], axis=-4).type)
        == "4 * (var * var * var * int64, var * var * var * int64)"
    )

    with pytest.raises(ValueError):
        ak.operations.cartesian([cuda_one, cuda_two], axis=-5)


def test_0193_is_none_axis_parameter():
    array = ak.Array([1, 2, 3, None, 5])
    cuda_array = ak.to_backend(array, "cuda")

    assert ak.operations.is_none(cuda_array).to_list() == [
        False,
        False,
        False,
        True,
        False,
    ]

    array = ak.Array([[1, 2, 3], [], [None, 5]])
    cuda_array = ak.to_backend(array, "cuda")

    assert ak.operations.is_none(cuda_array).to_list() == [
        False,
        False,
        False,
    ]
    assert ak.operations.is_none(cuda_array, axis=1).to_list() == [
        [False, False, False],
        [],
        [True, False],
    ]

    array = ak.Array([[1, None, 2, 3], [], [None, 5]])
    cuda_array = ak.to_backend(array, "cuda")

    assert ak.operations.is_none(cuda_array, axis=1).to_list() == [
        [False, True, False, False],
        [],
        [True, False],
    ]
    assert ak.operations.is_none(cuda_array, axis=-1).to_list() == [
        [False, True, False, False],
        [],
        [True, False],
    ]
    assert ak.operations.is_none(cuda_array, axis=-2).to_list() == [
        False,
        False,
        False,
    ]
    with pytest.raises(ValueError):
        ak.operations.is_none(cuda_array, axis=-3)

    array = ak.Array([[1, None, 2, 3], None, [None, 5]])
    cuda_array = ak.to_backend(array, "cuda")

    assert ak.operations.is_none(cuda_array, axis=-2).to_list() == [False, True, False]

    one = ak.operations.from_iter([1, None, 3], highlevel=False)
    two = ak.operations.from_iter([[], [1], None, [3, 3, 3]], highlevel=False)
    tags = ak.index.Index8(np.array([0, 1, 1, 0, 0, 1, 1], dtype=np.int8))
    index = ak.index.Index64(np.array([0, 0, 1, 1, 2, 2, 3], dtype=np.int64))
    array = ak.Array(ak.contents.UnionArray(tags, index, [one, two]), check_valid=True)
    cuda_array = ak.to_backend(array, "cuda")

    assert ak.to_list(cuda_array) == [1, [], [1], None, 3, None, [3, 3, 3]]

    assert ak.to_list(ak.operations.is_none(cuda_array)) == [
        False,
        False,
        False,
        True,
        False,
        True,
        False,
    ]


def test_0493_zeros_ones_full_like():
    array = ak.Array(
        [
            [{"x": 0.0, "y": []}, {"x": 1.1, "y": [1]}, {"x": 2.2, "y": [1, 2]}],
            [],
            [
                {"x": 3.3, "y": [1, 2, None, 3]},
                False,
                False,
                True,
                {"x": 4.4, "y": [1, 2, None, 3, 4]},
            ],
        ]
    )
    cuda_array = ak.to_backend(array, "cuda")

    assert ak.operations.full_like(cuda_array, 12.3).to_list() == [
        [{"x": 12.3, "y": []}, {"x": 12.3, "y": [12]}, {"x": 12.3, "y": [12, 12]}],
        [],
        [
            {"x": 12.3, "y": [12, 12, None, 12]},
            True,
            True,
            True,
            {"x": 12.3, "y": [12, 12, None, 12, 12]},
        ],
    ]

    assert ak.operations.zeros_like(cuda_array).to_list() == [
        [{"x": 0.0, "y": []}, {"x": 0.0, "y": [0]}, {"x": 0.0, "y": [0, 0]}],
        [],
        [
            {"x": 0.0, "y": [0, 0, None, 0]},
            False,
            False,
            False,
            {"x": 0.0, "y": [0, 0, None, 0, 0]},
        ],
    ]

    assert ak.operations.ones_like(cuda_array).to_list() == [
        [{"x": 1.0, "y": []}, {"x": 1.0, "y": [1]}, {"x": 1.0, "y": [1, 1]}],
        [],
        [
            {"x": 1.0, "y": [1, 1, None, 1]},
            True,
            True,
            True,
            {"x": 1.0, "y": [1, 1, None, 1, 1]},
        ],
    ]


def test_0496_provide_local_index():
    array = ak.highlevel.Array(
        [[[0.0, 1.1, 2.2], []], [[3.3, 4.4]], [], [[5.5], [], [6.6, 7.7, 8.8, 9.9]]]
    )
    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(ak.operations.local_index(cuda_array, axis=0)) == [
        0,
        1,
        2,
        3,
    ]
    assert to_list(ak.operations.local_index(cuda_array, axis=1)) == [
        [0, 1],
        [0],
        [],
        [0, 1, 2],
    ]
    assert to_list(ak.operations.local_index(cuda_array, axis=2)) == [
        [[0, 1, 2], []],
        [[0, 1]],
        [],
        [[0], [], [0, 1, 2, 3]],
    ]
    assert to_list(ak.operations.local_index(cuda_array, axis=-1)) == [
        [[0, 1, 2], []],
        [[0, 1]],
        [],
        [[0], [], [0, 1, 2, 3]],
    ]
    assert to_list(ak.operations.local_index(cuda_array, axis=-2)) == [
        [0, 1],
        [0],
        [],
        [0, 1, 2],
    ]
    assert to_list(ak.operations.local_index(cuda_array, axis=-3)) == [
        0,
        1,
        2,
        3,
    ]

    assert to_list(
        ak.operations.zip(
            [
                ak.operations.local_index(cuda_array, axis=0),
                ak.operations.local_index(cuda_array, axis=1),
                ak.operations.local_index(cuda_array, axis=2),
            ]
        )
    ) == [
        [[(0, 0, 0), (0, 0, 1), (0, 0, 2)], []],
        [[(1, 0, 0), (1, 0, 1)]],
        [],
        [[(3, 0, 0)], [], [(3, 2, 0), (3, 2, 1), (3, 2, 2), (3, 2, 3)]],
    ]


def test_0504_block_ufuncs_for_strings():
    def _apply_ufunc(ufunc, method, inputs, kwargs):
        nextinputs = []
        for x in inputs:
            if (
                isinstance(x, ak.highlevel.Array)
                and x.layout.is_indexed
                and not x.layout.is_option
            ):
                nextinputs.append(
                    ak.highlevel.Array(x.layout.project(), behavior=behavior_of(x))
                )
            else:
                nextinputs.append(x)

        return getattr(ufunc, method)(*nextinputs, **kwargs)

    behavior = {}
    behavior[np.ufunc, "categorical"] = _apply_ufunc

    array = ak.highlevel.Array(
        ak.contents.IndexedArray(
            ak.index.Index64(np.array([0, 1, 2, 1, 3, 1, 4])),
            ak.contents.NumpyArray(np.array([321, 1.1, 123, 999, 2])),
            parameters={"__array__": "categorical"},
        ),
        behavior=behavior,
    )
    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(cuda_array * 10) == [3210, 11, 1230, 11, 9990, 11, 20]

    cuda_array = ak.highlevel.Array(["HAL"])

    with pytest.raises(TypeError):
        cuda_array + 1


def test_0527_fix_unionarray_ufuncs_and_parameters_in_merging_0459():
    plain_plain = ak.highlevel.Array([[0.0, 1.1, 2.2, 3.3, 4.4]])
    cuda_plain_plain = ak.to_backend(plain_plain, "cuda")

    cuda_array_plain = ak.operations.with_parameter(
        cuda_plain_plain, "__list__", "zoinks"
    )
    cuda_plain_isdoc = ak.operations.with_parameter(
        cuda_plain_plain, "__doc__", "This is a zoink."
    )
    cuda_array_isdoc = ak.operations.with_parameter(
        cuda_array_plain, "__doc__", "This is a zoink."
    )

    assert ak.operations.parameters(cuda_plain_plain) == {}
    assert ak.operations.parameters(cuda_array_plain) == {"__list__": "zoinks"}
    assert ak.operations.parameters(cuda_plain_isdoc) == {"__doc__": "This is a zoink."}
    assert ak.operations.parameters(cuda_array_isdoc) == {
        "__list__": "zoinks",
        "__doc__": "This is a zoink.",
    }

    assert (
        ak.operations.parameters(
            ak.operations.concatenate([cuda_plain_plain, cuda_plain_plain])
        )
        == {}
    )
    assert ak.operations.parameters(
        ak.operations.concatenate([cuda_array_plain, cuda_array_plain])
    ) == {"__list__": "zoinks"}
    assert ak.operations.parameters(
        ak.operations.concatenate([cuda_plain_isdoc, cuda_plain_isdoc])
    ) == {"__doc__": "This is a zoink."}
    assert ak.operations.parameters(
        ak.operations.concatenate([cuda_array_isdoc, cuda_array_isdoc])
    ) == {
        "__list__": "zoinks",
        "__doc__": "This is a zoink.",
    }

    assert isinstance(
        ak.operations.concatenate([cuda_plain_plain, cuda_plain_plain]).layout,
        ak.contents.ListOffsetArray,
    )
    assert isinstance(
        ak.operations.concatenate([cuda_array_plain, cuda_array_plain]).layout,
        ak.contents.ListOffsetArray,
    )
    assert isinstance(
        ak.operations.concatenate([cuda_plain_isdoc, cuda_plain_isdoc]).layout,
        ak.contents.ListOffsetArray,
    )
    assert isinstance(
        ak.operations.concatenate([cuda_array_isdoc, cuda_array_isdoc]).layout,
        ak.contents.ListOffsetArray,
    )

    assert (
        ak.operations.parameters(
            ak.operations.concatenate([cuda_plain_plain, cuda_array_plain])
        )
        == {}
    )
    assert (
        ak.operations.parameters(
            ak.operations.concatenate([cuda_plain_isdoc, cuda_array_isdoc])
        )
        == {}
    )
    assert (
        ak.operations.parameters(
            ak.operations.concatenate([cuda_array_plain, cuda_plain_plain])
        )
        == {}
    )
    assert (
        ak.operations.parameters(
            ak.operations.concatenate([cuda_array_isdoc, cuda_plain_isdoc])
        )
        == {}
    )

    assert isinstance(
        ak.operations.concatenate([cuda_plain_plain, cuda_array_plain]).layout,
        ak.contents.UnionArray,
    )
    assert isinstance(
        ak.operations.concatenate([cuda_plain_isdoc, cuda_array_isdoc]).layout,
        ak.contents.UnionArray,
    )
    assert isinstance(
        ak.operations.concatenate([cuda_array_plain, cuda_plain_plain]).layout,
        ak.contents.UnionArray,
    )
    assert isinstance(
        ak.operations.concatenate([cuda_array_isdoc, cuda_plain_isdoc]).layout,
        ak.contents.UnionArray,
    )

    assert (
        ak.operations.parameters(
            ak.operations.concatenate([cuda_plain_plain, cuda_plain_isdoc])
        )
        == {}
    )
    assert ak.operations.parameters(
        ak.operations.concatenate([cuda_array_plain, cuda_array_isdoc])
    ) == {"__list__": "zoinks"}
    assert (
        ak.operations.parameters(
            ak.operations.concatenate([cuda_plain_isdoc, cuda_plain_plain])
        )
        == {}
    )
    assert ak.operations.parameters(
        ak.operations.concatenate([cuda_array_isdoc, cuda_array_plain])
    ) == {"__list__": "zoinks"}

    assert isinstance(
        ak.operations.concatenate([cuda_plain_plain, cuda_plain_isdoc]).layout,
        ak.contents.ListOffsetArray,
    )
    assert isinstance(
        ak.operations.concatenate([cuda_array_plain, cuda_array_isdoc]).layout,
        ak.contents.ListOffsetArray,
    )
    assert isinstance(
        ak.operations.concatenate([cuda_plain_isdoc, cuda_plain_plain]).layout,
        ak.contents.ListOffsetArray,
    )
    assert isinstance(
        ak.operations.concatenate([cuda_array_isdoc, cuda_array_plain]).layout,
        ak.contents.ListOffsetArray,
    )


def test_0527_fix_unionarray_ufuncs_and_parameters_in_merging_0522():
    content1 = ak.highlevel.Array([0.0, 1.1, 2.2, 3.3, 4.4]).layout
    content2 = ak.highlevel.Array([[0], [100], [200], [300], [400]]).layout
    tags = ak.index.Index8(np.array([0, 0, 0, 1, 1, 0, 0, 1, 1, 1], np.int8))
    index = ak.index.Index64(np.array([0, 1, 2, 0, 1, 3, 4, 2, 3, 4], np.int64))
    unionarray = ak.highlevel.Array(
        ak.contents.UnionArray(tags, index, [content1, content2])
    )
    cuda_unionarray = ak.to_backend(unionarray, "cuda")

    assert cuda_unionarray.to_list() == [
        0.0,
        1.1,
        2.2,
        [0],
        [100],
        3.3,
        4.4,
        [200],
        [300],
        [400],
    ]

    assert (cuda_unionarray + 10).to_list() == [
        10.0,
        11.1,
        12.2,
        [10],
        [110],
        13.3,
        14.4,
        [210],
        [310],
        [410],
    ]
    assert (10 + cuda_unionarray).to_list() == [
        10.0,
        11.1,
        12.2,
        [10],
        [110],
        13.3,
        14.4,
        [210],
        [310],
        [410],
    ]

    assert (cuda_unionarray + range(0, 100, 10)).to_list() == [
        0.0,
        11.1,
        22.2,
        [30],
        [140],
        53.3,
        64.4,
        [270],
        [380],
        [490],
    ]
    assert (range(0, 100, 10) + cuda_unionarray).to_list() == [
        0.0,
        11.1,
        22.2,
        [30],
        [140],
        53.3,
        64.4,
        [270],
        [380],
        [490],
    ]

    assert (cuda_unionarray + cp.arange(0, 100, 10)).tolist() == [
        0.0,
        11.1,
        22.2,
        [30],
        [140],
        53.3,
        64.4,
        [270],
        [380],
        [490],
    ]
    assert (cp.arange(0, 100, 10) + cuda_unionarray).tolist() == [
        0.0,
        11.1,
        22.2,
        [30],
        [140],
        53.3,
        64.4,
        [270],
        [380],
        [490],
    ]

    assert (cuda_unionarray + ak.highlevel.Array(cp.arange(0, 100, 10))).tolist() == [
        0.0,
        11.1,
        22.2,
        [30],
        [140],
        53.3,
        64.4,
        [270],
        [380],
        [490],
    ]
    assert (ak.highlevel.Array(cp.arange(0, 100, 10)) + cuda_unionarray).tolist() == [
        0.0,
        11.1,
        22.2,
        [30],
        [140],
        53.3,
        64.4,
        [270],
        [380],
        [490],
    ]

    # assert (cuda_unionarray + cuda_unionarray).to_list() == [
    #     0.0,
    #     2.2,
    #     4.4,
    #     [0],
    #     [200],
    #     6.6,
    #     8.8,
    #     [400],
    #     [600],
    #     [800],
    # ]


def test_0959_getitem_array_implementation_UnionArray_NumpyArray():
    v2a = ak.contents.unionarray.UnionArray(
        ak.index.Index(np.array([1, 1, 0, 0, 1, 0, 1], np.int8)),
        ak.index.Index(np.array([4, 3, 0, 1, 2, 2, 4, 100], np.int64)),
        [
            ak.from_iter([[1], [2], [3]], highlevel=False),
            ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
    )
    cuda_v2a = ak.to_backend(v2a, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2a[cp.array([0, 1, 3], cp.int64)]
    assert to_list(cuda_resultv2) == [5.5, 4.4, [2]]
    assert (
        cuda_v2a.to_typetracer()[cp.array([0, 1, 3], cp.int64)].form
        == cuda_resultv2.form
    )


def test_0959_getitem_array_implementation_BitMaskedArray_NumpyArray():
    v2a = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=True,
        length=13,
        lsb_order=False,
    )
    cuda_v2a = ak.to_backend(v2a, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2a[cp.array([0, 1, 4], cp.int64)]
    assert to_list(cuda_resultv2) == [0.0, 1.0, None]
    assert (
        cuda_v2a.to_typetracer()[cp.array([0, 1, 4], cp.int64)].form
        == cuda_resultv2.form
    )

    v2b = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=False,
        length=13,
        lsb_order=False,
    )
    cuda_v2b = ak.to_backend(v2b, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2b[cp.array([0, 1, 4], cp.int64)]
    assert to_list(cuda_resultv2) == [0.0, 1.0, None]
    assert (
        cuda_v2b.to_typetracer()[cp.array([0, 1, 4], cp.int64)].form
        == cuda_resultv2.form
    )

    v2c = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=True,
        length=13,
        lsb_order=True,
    )
    cuda_v2c = ak.to_backend(v2c, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2c[cp.array([0, 1, 4], cp.int64)]
    assert to_list(cuda_resultv2) == [0.0, 1.0, None]
    assert (
        cuda_v2c.to_typetracer()[cp.array([0, 1, 4], cp.int64)].form
        == cuda_resultv2.form
    )

    v2d = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.numpyarray.NumpyArray(
            np.array(
                [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6]
            )
        ),
        valid_when=False,
        length=13,
        lsb_order=True,
    )
    cuda_v2d = ak.to_backend(v2d, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2d[cp.array([0, 1, 4], cp.int64)]
    assert to_list(cuda_resultv2) == [0.0, 1.0, None]
    assert (
        cuda_v2d.to_typetracer()[cp.array([0, 1, 4], cp.int64)].form
        == cuda_resultv2.form
    )


def test_0959_getitem_array_implementation_BitMaskedArray_RecordArray_NumpyArray():
    v2a = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        True,
                        True,
                        True,
                        True,
                        False,
                        False,
                        False,
                        False,
                        True,
                        False,
                        True,
                        False,
                        True,
                    ]
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=True,
        length=13,
        lsb_order=False,
    )
    cuda_v2a = ak.to_backend(v2a, "cuda", highlevel=False)
    indexa = ak.index.Index(cp.array([0, 1, 4], cp.int64))
    cuda_resultv2 = cuda_v2a._carry(indexa, False)
    assert to_list(cuda_resultv2) == [{"nest": 0.0}, {"nest": 1.0}, None]
    assert (
        cuda_v2a.to_typetracer()
        ._carry(indexa.to_nplike(TypeTracer.instance()), False)
        .form
        == cuda_resultv2.form
    )

    v2b = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=False,
        length=13,
        lsb_order=False,
    )
    cuda_v2b = ak.to_backend(v2b, "cuda", highlevel=False)

    indexb = ak.index.Index(cp.array([1, 1, 4], cp.int64))
    cuda_resultv2 = cuda_v2b._carry(indexb, False)
    assert to_list(cuda_resultv2) == [{"nest": 1.0}, {"nest": 1.0}, None]
    assert (
        cuda_v2b.to_typetracer()
        ._carry(indexb.to_nplike(TypeTracer.instance()), False)
        .form
        == cuda_resultv2.form
    )

    v2c = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        0,
                        0,
                        0,
                        0,
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                        1,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=True,
        length=13,
        lsb_order=True,
    )
    cuda_v2c = ak.to_backend(v2c, "cuda", highlevel=False)

    indexc = ak.index.Index(cp.array([0, 1, 4], cp.int64))
    cuda_resultv2 = cuda_v2c._carry(indexc, False)
    assert to_list(cuda_resultv2) == [{"nest": 0.0}, {"nest": 1.0}, None]
    assert (
        cuda_v2c.to_typetracer()
        ._carry(indexc.to_nplike(TypeTracer.instance()), False)
        .form
        == cuda_resultv2.form
    )

    v2d = ak.contents.bitmaskedarray.BitMaskedArray(
        ak.index.Index(
            np.packbits(
                np.array(
                    [
                        1,
                        1,
                        1,
                        1,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        0,
                        1,
                        0,
                        1,
                        0,
                    ],
                    np.uint8,
                )
            )
        ),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array(
                        [
                            0.0,
                            1.0,
                            2.0,
                            3.0,
                            4.0,
                            5.0,
                            6.0,
                            7.0,
                            1.1,
                            2.2,
                            3.3,
                            4.4,
                            5.5,
                            6.6,
                        ]
                    )
                )
            ],
            ["nest"],
        ),
        valid_when=False,
        length=13,
        lsb_order=True,
    )
    cuda_v2d = ak.to_backend(v2d, "cuda", highlevel=False)

    indexd = ak.index.Index(cp.array([0, 0, 0], cp.int64))
    cuda_resultv2 = cuda_v2d._carry(indexd, False)
    assert to_list(cuda_resultv2) == [{"nest": 0.0}, {"nest": 0.0}, {"nest": 0.0}]
    assert (
        cuda_v2d.to_typetracer()
        ._carry(indexd.to_nplike(TypeTracer.instance()), False)
        .form
        == cuda_resultv2.form
    )


def test_0959_getitem_array_implementation_IndexedArray_RecordArray_NumpyArray():
    v2a = ak.contents.indexedarray.IndexedArray(
        ak.index.Index(np.array([2, 2, 0, 1, 4, 5, 4], np.int64)),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
    )
    cuda_v2a = ak.to_backend(v2a, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2a[cp.array([0, 1, 4], cp.int64)]
    assert to_list(cuda_resultv2) == [{"nest": 3.3}, {"nest": 3.3}, {"nest": 5.5}]
    assert (
        cuda_v2a.to_typetracer()[cp.array([0, 1, 4], cp.int64)].form
        == cuda_resultv2.form
    )


def test_0959_getitem_array_implementation_IndexedOptionArray_RecordArray_NumpyArray():
    v2a = ak.contents.indexedoptionarray.IndexedOptionArray(
        ak.index.Index(np.array([2, 2, -1, 1, -1, 5, 4], np.int64)),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
    )
    cuda_v2a = ak.to_backend(v2a, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2a[cp.array([0, 1, 4], cp.int64)]
    assert to_list(cuda_resultv2) == [{"nest": 3.3}, {"nest": 3.3}, None]
    assert (
        cuda_v2a.to_typetracer()[cp.array([0, 1, 4], cp.int64)].form
        == cuda_resultv2.form
    )


def test_0959_getitem_array_implementation_ByteMaskedArray_RecordArray_NumpyArray():
    v2a = ak.contents.bytemaskedarray.ByteMaskedArray(
        ak.index.Index(np.array([1, 0, 1, 0, 1], np.int8)),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
        valid_when=True,
    )
    cuda_v2a = ak.to_backend(v2a, "cuda", highlevel=False)

    index = ak.index.Index(cp.array([0, 1, 4], cp.int64))
    cuda_resultv2 = cuda_v2a._carry(index, False)
    assert to_list(cuda_resultv2) == [{"nest": 1.1}, None, {"nest": 5.5}]
    assert (
        cuda_v2a.to_typetracer()
        ._carry(index.to_nplike(TypeTracer.instance()), False)
        .form
        == cuda_resultv2.form
    )

    v2b = ak.contents.bytemaskedarray.ByteMaskedArray(
        ak.index.Index(np.array([0, 1, 0, 1, 0], np.int8)),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
        valid_when=False,
    )
    cuda_v2b = ak.to_backend(v2b, "cuda", highlevel=False)

    indexb = ak.index.Index(cp.array([3, 1, 4], cp.int64))
    cuda_resultv2 = cuda_v2b._carry(indexb, False)
    assert to_list(cuda_resultv2) == [None, None, {"nest": 5.5}]
    assert (
        cuda_v2b.to_typetracer()
        ._carry(indexb.to_nplike(TypeTracer.instance()), False)
        .form
        == cuda_resultv2.form
    )


def test_0959_getitem_array_implementation_IndexedArray_NumpyArray():
    v2a = ak.contents.indexedarray.IndexedArray(
        ak.index.Index(np.array([2, 2, 0, 1, 4, 5, 4], np.int64)),
        ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
    )
    cuda_v2a = ak.to_backend(v2a, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2a[cp.array([0, 1, 4], cp.int64)]
    assert to_list(cuda_resultv2) == [3.3, 3.3, 5.5]
    assert (
        cuda_v2a.to_typetracer()[cp.array([0, 1, 4], cp.int64)].form
        == cuda_resultv2.form
    )


def test_0959_getitem_array_implementation_IndexedOptionArray_NumpyArray():
    v2a = ak.contents.indexedoptionarray.IndexedOptionArray(
        ak.index.Index(np.array([2, 2, -1, 1, -1, 5, 4], np.int64)),
        ak.contents.numpyarray.NumpyArray(np.array([1.1, 2.2, 3.3, 4.4, 5.5, 6.6])),
    )
    cuda_v2a = ak.to_backend(v2a, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2a[cp.array([0, 1, -1], cp.int64)]
    assert to_list(cuda_resultv2) == [3.3, 3.3, 5.5]
    assert (
        cuda_v2a.to_typetracer()[cp.array([0, 1, -1], cp.int64)].form
        == cuda_resultv2.form
    )


def test_0959_getitem_array_implementation_RecordArray_NumpyArray():
    v2a = ak.contents.recordarray.RecordArray(
        [
            ak.contents.numpyarray.NumpyArray(np.array([0, 1, 2, 3, 4], np.int64)),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
        ["x", "y"],
    )
    cuda_v2a = ak.to_backend(v2a, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2a[cp.array([1, 2], cp.int64)]
    assert to_list(cuda_resultv2) == [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}]
    assert (
        cuda_v2a.to_typetracer()[cp.array([1, 2], cp.int64)].form == cuda_resultv2.form
    )

    v2b = ak.contents.recordarray.RecordArray(
        [
            ak.contents.numpyarray.NumpyArray(np.array([0, 1, 2, 3, 4], np.int64)),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
        None,
    )
    cuda_v2b = ak.to_backend(v2b, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2b[cp.array([0, 1, 2, 3, -1], cp.int64)]
    assert to_list(cuda_resultv2) == [(0, 0.0), (1, 1.1), (2, 2.2), (3, 3.3), (4, 4.4)]
    assert (
        cuda_v2b.to_typetracer()[cp.array([0, 1, 2, 3, -1], cp.int64)].form
        == cuda_resultv2.form
    )

    v2c = ak.contents.recordarray.RecordArray([], [], 10)
    cuda_v2c = ak.to_backend(v2c, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2c[cp.array([0], cp.int64)]
    assert to_list(cuda_resultv2) == [{}]
    assert cuda_v2c.to_typetracer()[cp.array([0], cp.int64)].form == cuda_resultv2.form

    v2d = ak.contents.recordarray.RecordArray([], None, 10)
    cuda_v2d = ak.to_backend(v2d, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2d[cp.array([0], cp.int64)]
    assert to_list(cuda_resultv2) == [()]
    assert cuda_v2d.to_typetracer()[cp.array([0], cp.int64)].form == cuda_resultv2.form


def test_0959_getitem_array_implementation_RegularArray_NumpyArray():
    v2a = ak.contents.regulararray.RegularArray(
        ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        3,
    )
    cuda_v2a = ak.to_backend(v2a, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2a[cp.array([0, 1], cp.int64)]
    assert to_list(cuda_resultv2) == [[0.0, 1.1, 2.2], [3.3, 4.4, 5.5]]
    assert (
        cuda_v2a.to_typetracer()[cp.array([0, 1], cp.int64)].form == cuda_resultv2.form
    )

    v2b = ak.contents.regulararray.RegularArray(
        ak.contents.emptyarray.EmptyArray(), 0, zeros_length=10
    )
    cuda_v2b = ak.to_backend(v2b, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2b[cp.array([0, 0, 0], cp.int64)]
    assert to_list(cuda_resultv2) == [[], [], []]
    assert (
        cuda_v2b.to_typetracer()[cp.array([0, 0, 0], cp.int64)].form
        == cuda_resultv2.form
    )

    assert to_list(cuda_resultv2) == [[], [], []]


def test_0959_getitem_array_implementation_ListArray_NumpyArray():
    v2a = ak.contents.listarray.ListArray(
        ak.index.Index(np.array([4, 100, 1], np.int64)),
        ak.index.Index(np.array([7, 100, 3, 200], np.int64)),
        ak.contents.numpyarray.NumpyArray(
            np.array([6.6, 4.4, 5.5, 7.7, 1.1, 2.2, 3.3, 8.8])
        ),
    )
    cuda_v2a = ak.to_backend(v2a, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2a[cp.array([1, -1], cp.int64)]
    assert to_list(cuda_resultv2) == [[], [4.4, 5.5]]
    assert (
        cuda_v2a.to_typetracer()[cp.array([1, -1], cp.int64)].form == cuda_resultv2.form
    )


def test_0959_getitem_array_implementation_UnionArray_RecordArray_NumpyArray():
    v2a = ak.contents.unionarray.UnionArray(
        ak.index.Index(np.array([1, 1, 0, 0, 1, 0, 1], np.int8)),
        ak.index.Index(np.array([4, 3, 0, 1, 2, 2, 4, 100], np.int64)),
        [
            ak.contents.recordarray.RecordArray(
                [ak.from_iter([[1], [2], [3]], highlevel=False)],
                ["nest"],
            ),
            ak.contents.recordarray.RecordArray(
                [
                    ak.contents.numpyarray.NumpyArray(
                        np.array([1.1, 2.2, 3.3, 4.4, 5.5])
                    )
                ],
                ["nest"],
            ),
        ],
    )
    cuda_v2a = ak.to_backend(v2a, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2a[cp.array([0, 1, 1], cp.int64)]
    assert to_list(cuda_resultv2) == [{"nest": 5.5}, {"nest": 4.4}, {"nest": 4.4}]
    assert (
        cuda_v2a.to_typetracer()[cp.array([0, 1, 1], cp.int64)].form
        == cuda_resultv2.form
    )


def test_0959_getitem_array_implementation_RecordArray_NumpyArray_lazy():
    v2a = ak.contents.recordarray.RecordArray(
        [
            ak.contents.numpyarray.NumpyArray(np.array([0, 1, 2, 3, 4], np.int64)),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
        ["x", "y"],
    )
    cuda_v2a = ak.to_backend(v2a, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2a._carry(ak.index.Index(cp.array([1, 2], cp.int64)), True)
    assert to_list(cuda_resultv2) == [{"x": 1, "y": 1.1}, {"x": 2, "y": 2.2}]
    assert (
        cuda_v2a.to_typetracer()
        ._carry(ak.index.Index(cp.array([1, 2], cp.int64)), True)
        .form
        == cuda_resultv2.form
    )

    v2b = ak.contents.recordarray.RecordArray(
        [
            ak.contents.numpyarray.NumpyArray(np.array([0, 1, 2, 3, 4], np.int64)),
            ak.contents.numpyarray.NumpyArray(np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5])),
        ],
        None,
    )
    cuda_v2b = ak.to_backend(v2b, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2b._carry(
        ak.index.Index(cp.array([0, 1, 2, 3, 4], cp.int64)), True
    )
    assert to_list(cuda_resultv2) == [(0, 0.0), (1, 1.1), (2, 2.2), (3, 3.3), (4, 4.4)]
    assert (
        cuda_v2b.to_typetracer()
        ._carry(ak.index.Index(cp.array([0, 1, 2, 3, 4], cp.int64)), True)
        .form
        == cuda_resultv2.form
    )

    v2c = ak.contents.recordarray.RecordArray([], [], 10)
    cuda_v2c = ak.to_backend(v2c, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2c[cp.array([0], cp.int64)]
    assert to_list(cuda_resultv2) == [{}]
    assert cuda_v2c.to_typetracer()[cp.array([0], cp.int64)].form == cuda_resultv2.form

    v2d = ak.contents.recordarray.RecordArray([], None, 10)
    cuda_v2d = ak.to_backend(v2d, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2d[cp.array([0], cp.int64)]
    assert to_list(cuda_resultv2) == [()]
    assert cuda_v2d.to_typetracer()[cp.array([0], cp.int64)].form == cuda_resultv2.form


def test_0959_getitem_array_implementation_RegularArray_RecordArray_NumpyArray():
    v2a = ak.contents.regulararray.RegularArray(
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6])
                )
            ],
            ["nest"],
        ),
        3,
    )
    cuda_v2a = ak.to_backend(v2a, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2a._carry(ak.index.Index(cp.array([0], cp.int64)), False)
    assert to_list(cuda_resultv2) == [[{"nest": 0.0}, {"nest": 1.1}, {"nest": 2.2}]]
    assert (
        cuda_v2a.to_typetracer()
        ._carry(ak.index.Index(cp.array([0], cp.int64)), False)
        .form
        == cuda_resultv2.form
    )

    v2b = ak.contents.regulararray.RegularArray(
        ak.contents.recordarray.RecordArray(
            [ak.contents.emptyarray.EmptyArray()], ["nest"]
        ),
        0,
        zeros_length=10,
    )
    cuda_v2b = ak.to_backend(v2b, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2b._carry(ak.index.Index(cp.array([0], cp.int64)), False)
    assert to_list(cuda_resultv2) == [[]]
    assert (
        cuda_v2b.to_typetracer()
        ._carry(ak.index.Index(cp.array([0], cp.int64)), False)
        .form
        == cuda_resultv2.form
    )


def test_0959_getitem_array_implementation_ListArray_RecordArray_NumpyArray():
    v2a = ak.contents.listarray.ListArray(
        ak.index.Index(np.array([4, 100, 1], np.int64)),
        ak.index.Index(np.array([7, 100, 3, 200], np.int64)),
        ak.contents.recordarray.RecordArray(
            [
                ak.contents.numpyarray.NumpyArray(
                    np.array([6.6, 4.4, 5.5, 7.7, 1.1, 2.2, 3.3, 8.8])
                )
            ],
            ["nest"],
        ),
    )
    cuda_v2a = ak.to_backend(v2a, "cuda", highlevel=False)

    cuda_resultv2 = cuda_v2a[cp.array([0, 1], np.int64)]
    assert to_list(cuda_resultv2) == [[{"nest": 1.1}, {"nest": 2.2}, {"nest": 3.3}], []]
    assert (
        cuda_v2a.to_typetracer()[np.array([0, 1], np.int64)].form == cuda_resultv2.form
    )



# BSD 3-Clause License; see https://github.com/scikit-hep/awkward/blob/main/LICENSE

from __future__ import annotations

from functools import cached_property

from awkward._meta.meta import Meta
from awkward._regularize import is_integer
from awkward._typing import Generic, JSONSerializable, TypeVar
from awkward.errors import FieldNotFoundError

T = TypeVar("T", bound=Meta)


class RecordMeta(Meta, Generic[T]):
    is_record = True

    _contents: list[T]
    _fields: list[str] | None

    @property
    def is_tuple(self) -> bool:
        return self._fields is None

    def purelist_parameters(self, *keys: str) -> JSONSerializable:
        if self._parameters is not None:
            for key in keys:
                if key in self._parameters:
                    return self._parameters[key]
        return None

    @property
    def purelist_isregular(self) -> bool:
        return True

    @property
    def purelist_depth(self) -> int:
        return 1

    @property
    def is_identity_like(self) -> bool:
        return False

    @cached_property
    def minmax_depth(self) -> tuple[int, int]:  # type: ignore[override]
        if len(self._contents) == 0:
            return (1, 1)
        mindepth, maxdepth = self._contents[0].minmax_depth
        for content in self._contents[1:]:
            mindepth_, maxdepth_ = content.minmax_depth
            if mindepth_ < mindepth:
                mindepth = mindepth_
            if maxdepth_ > maxdepth:
                maxdepth = maxdepth_
        return (mindepth, maxdepth)

    @cached_property
    def branch_depth(self) -> tuple[bool, int]:  # type: ignore[override]
        if len(self._contents) == 0:
            return False, 1

        any_branch = False
        min_depth = None
        for content in self._contents:
            branch, depth = content.branch_depth
            if min_depth is None:
                min_depth = depth
            if branch or min_depth != depth:
                any_branch = True
            if min_depth > depth:
                min_depth = depth
        assert min_depth is not None
        return any_branch, min_depth

    @property
    def dimension_optiontype(self) -> bool:
        return False

    @property
    def is_leaf(self) -> bool:  # type: ignore[override]
        return len(self._contents) == 0

    @property
    def contents(self) -> list[T]:
        return self._contents

    @cached_property
    def fields(self) -> list[str]:  # type: ignore[override]
        if self._fields is None:
            return [str(i) for i in range(len(self._contents))]
        else:
            return self._fields

    def index_to_field(self, index: int) -> str:
        if 0 <= index < len(self._contents):
            if self._fields is None:
                return str(index)
            else:
                return self._fields[index]
        else:
            raise IndexError(
                f"no index {index} in record with {len(self._contents)} fields"
            )

    def field_to_index(self, field: str) -> int:
        if self._fields is None:
            try:
                i = int(field)
            except ValueError:
                pass
            else:
                if 0 <= i < len(self._contents):
                    return i
        else:
            try:
                i = self._fields.index(field)
            except ValueError:
                pass
            else:
                return i
        raise FieldNotFoundError(
            f"no field {field!r} in record with {len(self._contents)} fields"
        )

    def has_field(self, field: str) -> bool:
        if self._fields is None:
            try:
                i = int(field)
            except ValueError:
                return False
            else:
                return 0 <= i < len(self._contents)
        else:
            return field in self._fields

    def content(self, index_or_field: int | str) -> T:
        if is_integer(index_or_field):
            index = index_or_field
        elif isinstance(index_or_field, str):
            index = self.field_to_index(index_or_field)
        else:
            raise TypeError(
                f"index_or_field must be an integer (index) or string (field), not {index_or_field!r}"
            )
        return self._contents[index]  # type: ignore[index]



from __future__ import annotations

import cupy as cp
import numpy as np
import pytest

import awkward as ak

to_list = ak.operations.to_list


@pytest.fixture(scope="function", autouse=True)
def cleanup_cuda():
    yield
    try:
        cp.cuda.Device().synchronize()  # wait for all kernels
    except cp.cuda.runtime.CUDARuntimeError as e:
        print("GPU error during sync:", e)
    cp._default_memory_pool.free_all_blocks()


def test_0111_jagged_and_masked_getitem_bitmaskedarray2b():
    array = ak.operations.from_iter(
        [[0.0, 1.1, 2.2], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )

    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    mask = ak.index.IndexU8(cp.array([66], dtype=cp.uint8))
    maskedarray = ak.contents.BitMaskedArray(
        mask, cuda_array, valid_when=False, length=4, lsb_order=True
    )  # lsb_order is irrelevant in this example
    cuda_array = ak.highlevel.Array([[1, 1], None, [0], [0, -1]], backend="cuda").layout

    assert to_list(maskedarray) == [
        [0.0, 1.1, 2.2],
        None,
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert to_list(maskedarray[cuda_array]) == [
        [1.1, 1.1],
        None,
        [5.5],
        [6.6, 9.9],
    ]
    assert maskedarray.to_typetracer()[cuda_array].form == maskedarray[cuda_array].form

    del cuda_array


def test_0111_jagged_and_masked_getitem_bytemaskedarray2b():
    array = ak.operations.from_iter(
        [[0.0, 1.1, 2.2], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )

    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    mask = ak.index.Index8(cp.array([0, 1, 0, 0], dtype=cp.int8))
    maskedarray = ak.contents.ByteMaskedArray(mask, cuda_array, valid_when=False)

    cuda_array = ak.highlevel.Array([[1, 1], None, [0], [0, -1]], backend="cuda").layout

    assert to_list(maskedarray) == [
        [0.0, 1.1, 2.2],
        None,
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert to_list(maskedarray[cuda_array]) == [
        [1.1, 1.1],
        None,
        [5.5],
        [6.6, 9.9],
    ]
    assert maskedarray.to_typetracer()[cuda_array].form == maskedarray[cuda_array].form
    del cuda_array


def test_0111_jagged_and_masked_getitem_emptyarray():
    content = ak.contents.EmptyArray()
    offsets = ak.index.Index64(np.array([0, 0, 0, 0, 0], dtype=np.int64))
    listoffsetarray = ak.contents.ListOffsetArray(offsets, content)

    cuda_listoffsetarray = ak.to_backend(listoffsetarray, "cuda", highlevel=False)

    array1 = ak.highlevel.Array([[], [], [], []], check_valid=True).layout
    cuda_array1 = ak.to_backend(array1, "cuda")

    array2 = ak.highlevel.Array([[], [None], [], []], check_valid=True).layout
    cuda_array2 = ak.to_backend(array2, "cuda")

    array3 = ak.highlevel.Array([[], [], None, []], check_valid=True).layout
    cuda_array3 = ak.to_backend(array3, "cuda")

    array4 = ak.highlevel.Array([[], [None], None, []], check_valid=True).layout
    cuda_array4 = ak.to_backend(array4, "cuda")

    array5 = ak.highlevel.Array([[], [0], [], []], check_valid=True).layout
    cuda_array5 = ak.to_backend(array5, "cuda")

    assert to_list(cuda_listoffsetarray) == [[], [], [], []]

    assert to_list(cuda_listoffsetarray[cuda_array1]) == [[], [], [], []]
    assert (
        cuda_listoffsetarray.to_typetracer()[cuda_array1].form
        == cuda_listoffsetarray[cuda_array1].form
    )

    assert to_list(cuda_listoffsetarray[cuda_array2]) == [[], [None], [], []]
    assert (
        cuda_listoffsetarray.to_typetracer()[cuda_array2].form
        == cuda_listoffsetarray[cuda_array2].form
    )
    assert to_list(cuda_listoffsetarray[cuda_array3]) == [[], [], None, []]
    assert (
        cuda_listoffsetarray.to_typetracer()[cuda_array3].form
        == cuda_listoffsetarray[cuda_array3].form
    )
    assert to_list(cuda_listoffsetarray[cuda_array4]) == [[], [None], None, []]
    assert (
        cuda_listoffsetarray.to_typetracer()[cuda_array4].form
        == cuda_listoffsetarray[cuda_array4].form
    )

    with pytest.raises(IndexError):
        cuda_listoffsetarray[cuda_array5]

    del cuda_listoffsetarray


def test_0111_jagged_and_masked_getitem_indexedarray():
    array = ak.operations.from_iter(
        [[0.0, 1.1, 2.2], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    index = ak.index.Index64(cp.array([3, 2, 1, 0], dtype=cp.int64))
    cuda_indexedarray = ak.contents.IndexedArray(index, cuda_array)

    assert to_list(cuda_indexedarray) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        [0.0, 1.1, 2.2],
    ]

    array1 = ak.highlevel.Array([[0, -1], [0], [], [1, 1]], check_valid=True).layout
    cuda_array1 = ak.to_backend(array1, "cuda")

    assert to_list(cuda_indexedarray[cuda_array1]) == [
        [6.6, 9.9],
        [5.5],
        [],
        [1.1, 1.1],
    ]
    assert (
        cuda_indexedarray.to_typetracer()[cuda_array1].form
        == cuda_indexedarray[cuda_array1].form
    )

    array1 = ak.highlevel.Array(
        [[0, -1], [0], [None], [1, None, 1]], check_valid=True
    ).layout
    cuda_array1 = ak.to_backend(array1, "cuda")

    assert to_list(cuda_indexedarray[cuda_array1]) == [
        [6.6, 9.9],
        [5.5],
        [None],
        [1.1, None, 1.1],
    ]
    assert (
        cuda_indexedarray.to_typetracer()[cuda_array1].form
        == cuda_indexedarray[cuda_array1].form
    )

    array1 = ak.highlevel.Array([[0, -1], [0], None, [1, 1]], check_valid=True).layout
    cuda_array1 = ak.to_backend(array1, "cuda")

    assert to_list(cuda_indexedarray[cuda_array1]) == [
        [6.6, 9.9],
        [5.5],
        None,
        [1.1, 1.1],
    ]
    assert (
        cuda_indexedarray.to_typetracer()[cuda_array1].form
        == cuda_indexedarray[cuda_array1].form
    )

    array1 = ak.highlevel.Array([[0, -1], [0], None, [None]], check_valid=True).layout
    cuda_array1 = ak.to_backend(array1, "cuda")

    assert to_list(cuda_indexedarray[cuda_array1]) == [[6.6, 9.9], [5.5], None, [None]]
    assert (
        cuda_indexedarray.to_typetracer()[cuda_array1].form
        == cuda_indexedarray[cuda_array1].form
    )

    index = ak.index.Index64(cp.array([3, 2, 1, 0], dtype=cp.int64))
    cuda_indexedarray = ak.contents.IndexedOptionArray(index, cuda_array)

    assert to_list(cuda_indexedarray) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        [3.3, 4.4],
        [0.0, 1.1, 2.2],
    ]

    array1 = ak.highlevel.Array([[0, -1], [0], [], [1, 1]], check_valid=True)
    cuda_array1 = ak.to_backend(array1, "cuda")

    assert to_list(cuda_indexedarray[cuda_array1]) == [
        [6.6, 9.9],
        [5.5],
        [],
        [1.1, 1.1],
    ]
    assert (
        cuda_indexedarray.to_typetracer()[cuda_array1].form
        == cuda_indexedarray[cuda_array1].form
    )

    array1 = ak.highlevel.Array(
        [[0, -1], [0], [None], [1, None, 1]], check_valid=True
    ).layout
    cuda_array1 = ak.to_backend(array1, "cuda")

    assert to_list(cuda_indexedarray[cuda_array1]) == [
        [6.6, 9.9],
        [5.5],
        [None],
        [1.1, None, 1.1],
    ]
    assert (
        cuda_indexedarray.to_typetracer()[cuda_array1].form
        == cuda_indexedarray[cuda_array1].form
    )

    array1 = ak.highlevel.Array([[0, -1], [0], None, []], check_valid=True).layout
    cuda_array1 = ak.to_backend(array1, "cuda")

    assert to_list(cuda_indexedarray[cuda_array1]) == [[6.6, 9.9], [5.5], None, []]
    assert (
        cuda_indexedarray.to_typetracer()[array1].form
        == cuda_indexedarray[cuda_array1].form
    )

    array1 = ak.highlevel.Array(
        [[0, -1], [0], None, [1, None, 1]], check_valid=True
    ).layout
    cuda_array1 = ak.to_backend(array1, "cuda")

    assert to_list(cuda_indexedarray[cuda_array1]) == [
        [6.6, 9.9],
        [5.5],
        None,
        [1.1, None, 1.1],
    ]
    assert (
        cuda_indexedarray.to_typetracer()[cuda_array1].form
        == cuda_indexedarray[cuda_array1].form
    )

    del cuda_indexedarray
    del cuda_array1


def test_0111_jagged_and_masked_getitem_indexedarray2():
    array = ak.operations.from_iter(
        [[0.0, 1.1, 2.2], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    index = ak.index.Index64(cp.array([3, 2, -1, 0], dtype=cp.int64))
    cuda_indexedarray = ak.contents.IndexedOptionArray(index, cuda_array)
    cuda_array = ak.highlevel.Array([[0, -1], [0], None, [1, 1]], backend="cuda").layout

    assert to_list(cuda_indexedarray) == [
        [6.6, 7.7, 8.8, 9.9],
        [5.5],
        None,
        [0.0, 1.1, 2.2],
    ]
    assert to_list(cuda_indexedarray[cuda_array]) == [
        [6.6, 9.9],
        [5.5],
        None,
        [1.1, 1.1],
    ]
    assert (
        cuda_indexedarray.to_typetracer()[cuda_array].form
        == cuda_indexedarray[cuda_array].form
    )
    del cuda_indexedarray
    del cuda_array


def test_0111_jagged_and_masked_getitem_indexedarray2b():
    array = ak.operations.from_iter(
        [[0.0, 1.1, 2.2], [3.3, 4.4], [5.5], [6.6, 7.7, 8.8, 9.9]], highlevel=False
    )
    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    index = ak.index.Index64(cp.array([0, -1, 2, 3], dtype=cp.int64))
    cuda_indexedarray = ak.contents.IndexedOptionArray(index, cuda_array)
    cuda_array = ak.highlevel.Array([[1, 1], None, [0], [0, -1]], backend="cuda").layout

    assert to_list(cuda_indexedarray) == [
        [0.0, 1.1, 2.2],
        None,
        [5.5],
        [6.6, 7.7, 8.8, 9.9],
    ]
    assert to_list(cuda_indexedarray[cuda_array]) == [
        [1.1, 1.1],
        None,
        [5.5],
        [6.6, 9.9],
    ]
    assert (
        cuda_indexedarray.to_typetracer()[cuda_array].form
        == cuda_indexedarray[cuda_array].form
    )
    del cuda_indexedarray
    del cuda_array


def test_0111_jagged_and_masked_getitem_indexedarray3():
    array = ak.highlevel.Array([0.0, 1.1, 2.2, None, 4.4, None, None, 7.7]).layout
    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(cuda_array[ak.highlevel.Array([4, 3, 2], backend="cuda")]) == [
        4.4,
        None,
        2.2,
    ]
    assert to_list(
        cuda_array[ak.highlevel.Array([4, 3, 2, None, 1], backend="cuda")]
    ) == [
        4.4,
        None,
        2.2,
        None,
        1.1,
    ]

    array = ak.highlevel.Array([[0.0, 1.1, None, 2.2], [3.3, None, 4.4], [5.5]]).layout
    array2 = ak.highlevel.Array([[3, 2, 2, 1], [1, 2], []]).layout

    cuda_array = ak.to_backend(array, "cuda", highlevel=False)
    cuda_array2 = ak.to_backend(array2, "cuda")

    assert to_list(cuda_array[cuda_array2]) == [
        [2.2, None, None, 1.1],
        [None, 4.4],
        [],
    ]
    assert cuda_array.to_typetracer()[cuda_array2].form == cuda_array[cuda_array2].form

    cuda_array = ak.highlevel.Array(
        [[0.0, 1.1, 2.2], [3.3, 4.4], None, [5.5]], backend="cuda"
    ).layout
    cuda_array2 = ak.highlevel.Array([3, 2, 1], backend="cuda").layout
    cuda_array3 = ak.highlevel.Array([3, 2, 1, None, 0], backend="cuda").layout
    cuda_array4 = ak.highlevel.Array(
        [[2, 1, 1, 0], [1], None, [0]], backend="cuda"
    ).layout
    cuda_array5 = ak.highlevel.Array(
        [[2, 1, 1, 0], None, [1], [0]], backend="cuda"
    ).layout
    cuda_array6 = ak.highlevel.Array(
        [[2, 1, 1, 0], None, [1], [0], None], backend="cuda"
    ).layout

    assert to_list(cuda_array[cuda_array2]) == [[5.5], None, [3.3, 4.4]]
    assert cuda_array.to_typetracer()[cuda_array2].form == cuda_array[cuda_array2].form
    assert to_list(cuda_array[cuda_array3]) == [
        [5.5],
        None,
        [3.3, 4.4],
        None,
        [0.0, 1.1, 2.2],
    ]
    assert cuda_array.to_typetracer()[cuda_array3].form == cuda_array[cuda_array3].form

    assert (to_list(cuda_array[cuda_array4])) == [
        [2.2, 1.1, 1.1, 0.0],
        [4.4],
        None,
        [5.5],
    ]
    assert cuda_array.to_typetracer()[cuda_array4].form == cuda_array[cuda_array4].form

    assert to_list(cuda_array[cuda_array5]) == [
        [2.2, 1.1, 1.1, 0],
        None,
        None,
        [5.5],
    ]
    assert cuda_array.to_typetracer()[cuda_array5].form == cuda_array[cuda_array5].form
    with pytest.raises(IndexError):
        cuda_array[cuda_array6]

    del cuda_array
    del cuda_array2
    del cuda_array3
    del cuda_array4
    del cuda_array5
    del cuda_array6


def test_0111_jagged_and_masked_getitem_jagged():
    array = ak.highlevel.Array(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]], check_valid=True
    ).layout
    array2 = ak.highlevel.Array(
        [[0, -1], [], [-1, 0], [-1], [1, 1, -2, 0]], check_valid=True
    ).layout

    cuda_array = ak.to_backend(array, "cuda", highlevel=False)
    cuda_array2 = ak.to_backend(array2, "cuda")

    assert to_list(cuda_array[cuda_array2]) == [
        [1.1, 3.3],
        [],
        [5.5, 4.4],
        [6.6],
        [8.8, 8.8, 8.8, 7.7],
    ]
    assert cuda_array.to_typetracer()[cuda_array2].form == cuda_array[cuda_array2].form

    del cuda_array
    del cuda_array2


def test_0111_jagged_and_masked_getitem_double_jagged():
    array = ak.highlevel.Array(
        [[[0, 1, 2, 3], [4, 5]], [[6, 7, 8], [9, 10, 11, 12, 13]]], check_valid=True
    ).layout
    array2 = ak.highlevel.Array(
        [[[2, 1, 0], [-1]], [[-1, -2, -3], [2, 1, 1, 3]]], check_valid=True
    ).layout

    cuda_array = ak.to_backend(array, "cuda", highlevel=False)
    cuda_array2 = ak.to_backend(array2, "cuda")

    assert to_list(cuda_array[cuda_array2]) == [
        [[2, 1, 0], [5]],
        [[8, 7, 6], [11, 10, 10, 12]],
    ]
    assert cuda_array.to_typetracer()[cuda_array2].form == cuda_array[cuda_array2].form

    content = ak.operations.from_iter(
        [[0, 1, 2, 3], [4, 5], [6, 7, 8], [9, 10, 11, 12, 13]], highlevel=False
    )
    regulararray = ak.contents.RegularArray(content, 2, zeros_length=0)
    cuda_regulararray = ak.to_backend(regulararray, "cuda", highlevel=False)

    array1 = ak.highlevel.Array([[2, 1, 0], [-1]], check_valid=True).layout
    cuda_array1 = ak.to_backend(array1, "cuda")

    assert to_list(cuda_regulararray[:, cuda_array1]) == [
        [[2, 1, 0], [5]],
        [[8, 7, 6], [13]],
    ]
    assert (
        cuda_regulararray.to_typetracer()[:, cuda_array1].form
        == cuda_regulararray[:, cuda_array1].form
    )
    assert to_list(cuda_regulararray[1:, cuda_array1]) == [[[8, 7, 6], [13]]]
    assert (
        cuda_regulararray.to_typetracer()[1:, cuda_array1].form
        == cuda_regulararray[1:, cuda_array1].form
    )

    offsets = ak.index.Index64(np.array([0, 2, 4], dtype=np.int64))
    listoffsetarray = ak.contents.ListOffsetArray(offsets, content)
    cuda_listoffsetarray = ak.to_backend(listoffsetarray, "cuda", highlevel=False)

    assert to_list(cuda_listoffsetarray[:, cuda_array1]) == [
        [[2, 1, 0], [5]],
        [[8, 7, 6], [13]],
    ]
    assert (
        cuda_listoffsetarray.to_typetracer()[:, cuda_array1].form
        == cuda_listoffsetarray[:, cuda_array1].form
    )
    assert to_list(cuda_listoffsetarray[1:, cuda_array1]) == [[[8, 7, 6], [13]]]
    assert (
        cuda_listoffsetarray.to_typetracer()[1:, cuda_array1].form
        == cuda_listoffsetarray[1:, cuda_array1].form
    )


def test_0111_jagged_and_masked_getitem_masked_jagged():
    array = ak.highlevel.Array(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]], check_valid=True
    ).layout
    array1 = ak.highlevel.Array(
        [[-1, -2], None, [], None, [-2, 0]], check_valid=True
    ).layout
    cuda_array = ak.to_backend(array, "cuda", highlevel=False)
    cuda_array1 = ak.to_backend(array1, "cuda")

    assert to_list(cuda_array[cuda_array1]) == [[3.3, 2.2], None, [], None, [8.8, 7.7]]
    assert cuda_array.to_typetracer()[cuda_array1].form == cuda_array[cuda_array1].form


def test_0111_jagged_and_masked_getitem_jagged_masked():
    array = ak.highlevel.Array(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]], check_valid=True
    ).layout
    array1 = ak.highlevel.Array(
        [[-1, None], [], [None, 0], [None], [1]], check_valid=True
    ).layout
    cuda_array = ak.to_backend(array, "cuda", highlevel=False)
    cuda_array1 = ak.to_backend(array1, "cuda")

    assert to_list(cuda_array[cuda_array1]) == [
        [3.3, None],
        [],
        [None, 4.4],
        [None],
        [8.8],
    ]
    assert cuda_array.to_typetracer()[cuda_array1].form == cuda_array[cuda_array1].form


def test_0111_jagged_and_masked_getitem_array_boolean_to_int():
    a = ak.operations.from_iter(
        [[True, True, True], [], [True, True], [True], [True, True, True, True]],
        highlevel=False,
    )
    cuda_a = ak.to_backend(a, "cuda", highlevel=False)
    b = ak._slicing._normalise_item_bool_to_int(cuda_a, backend=cuda_a.backend)
    assert to_list(b) == [[0, 1, 2], [], [0, 1], [0], [0, 1, 2, 3]]

    a = ak.operations.from_iter(
        [
            [True, True, False],
            [],
            [True, False],
            [False],
            [True, True, True, False],
        ],
        highlevel=False,
    )
    cuda_a = ak.to_backend(a, "cuda", highlevel=False)
    b = ak._slicing._normalise_item_bool_to_int(cuda_a, backend=cuda_a.backend)
    assert to_list(b) == [[0, 1], [], [0], [], [0, 1, 2]]

    a = ak.operations.from_iter(
        [
            [False, True, True],
            [],
            [False, True],
            [False],
            [False, True, True, True],
        ],
        highlevel=False,
    )
    cuda_a = ak.to_backend(a, "cuda", highlevel=False)
    b = ak._slicing._normalise_item_bool_to_int(cuda_a, backend=cuda_a.backend)
    assert to_list(b) == [[1, 2], [], [1], [], [1, 2, 3]]

    a = ak.operations.from_iter(
        [[True, True, None], [], [True, None], [None], [True, True, True, None]],
        highlevel=False,
    )
    cuda_a = ak.to_backend(a, "cuda", highlevel=False)
    b = ak._slicing._normalise_item_bool_to_int(cuda_a, backend=cuda_a.backend)
    assert to_list(b) == [[0, 1, None], [], [0, None], [None], [0, 1, 2, None]]
    assert (
        b.content.index.data[b.content.index.data >= 0].tolist()
        == np.arange(6).tolist()  # kernels expect nonnegative entries to be arange
    )

    a = ak.operations.from_iter(
        [[None, True, True], [], [None, True], [None], [None, True, True, True]],
        highlevel=False,
    )
    cuda_a = ak.to_backend(a, "cuda", highlevel=False)
    b = ak._slicing._normalise_item_bool_to_int(cuda_a, backend=cuda_a.backend)
    assert to_list(b) == [[None, 1, 2], [], [None, 1], [None], [None, 1, 2, 3]]
    assert (
        b.content.index.data[b.content.index.data >= 0].tolist()
        == np.arange(6).tolist()  # kernels expect nonnegative entries to be arange
    )

    a = ak.operations.from_iter(
        [[False, True, None], [], [False, None], [None], [False, True, True, None]],
        highlevel=False,
    )
    cuda_a = ak.to_backend(a, "cuda", highlevel=False)
    b = ak._slicing._normalise_item_bool_to_int(cuda_a, backend=cuda_a.backend)
    assert to_list(b) == [[1, None], [], [None], [None], [1, 2, None]]
    assert (
        b.content.index.data[b.content.index.data >= 0].tolist()
        == np.arange(3).tolist()  # kernels expect nonnegative entries to be arange
    )

    a = ak.operations.from_iter(
        [[None, True, False], [], [None, False], [None], [None, True, True, False]],
        highlevel=False,
    )
    cuda_a = ak.to_backend(a, "cuda", highlevel=False)
    b = ak._slicing._normalise_item_bool_to_int(cuda_a, backend=cuda_a.backend)
    assert to_list(b) == [[None, 1], [], [None], [None], [None, 1, 2]]
    assert (
        b.content.index.data[b.content.index.data >= 0].tolist()
        == np.arange(3).tolist()  # kernels expect nonnegative entries to be arange
    )


def test_0111_jagged_and_masked_getitem_array_slice():
    array = ak.highlevel.Array(
        [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], check_valid=True
    ).layout
    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    assert to_list(cuda_array[[5, 2, 2, 3, 9, 0, 1]]) == [
        5.5,
        2.2,
        2.2,
        3.3,
        9.9,
        0.0,
        1.1,
    ]
    assert (
        cuda_array.to_typetracer()[[5, 2, 2, 3, 9, 0, 1]].form
        == cuda_array[[5, 2, 2, 3, 9, 0, 1]].form
    )
    assert to_list(cuda_array[cp.array([5, 2, 2, 3, 9, 0, 1])]) == [
        5.5,
        2.2,
        2.2,
        3.3,
        9.9,
        0.0,
        1.1,
    ]
    assert (
        cuda_array.to_typetracer()[cp.array([5, 2, 2, 3, 9, 0, 1])].form
        == cuda_array[cp.array([5, 2, 2, 3, 9, 0, 1])].form
    )

    array2 = ak.contents.NumpyArray(np.array([5, 2, 2, 3, 9, 0, 1], dtype=np.int32))
    cuda_array2 = ak.to_backend(array2, "cuda")

    assert to_list(cuda_array[cuda_array2]) == [5.5, 2.2, 2.2, 3.3, 9.9, 0.0, 1.1]
    assert cuda_array.to_typetracer()[cuda_array2].form == cuda_array[cuda_array2].form
    assert to_list(
        cuda_array[
            ak.highlevel.Array(
                cp.array([5, 2, 2, 3, 9, 0, 1], dtype=cp.int32), check_valid=True
            )
        ]
    ) == [5.5, 2.2, 2.2, 3.3, 9.9, 0.0, 1.1]
    assert (
        cuda_array.to_typetracer()[
            ak.highlevel.Array(
                cp.array([5, 2, 2, 3, 9, 0, 1], dtype=cp.int32), check_valid=True
            )
        ].form
        == cuda_array[
            ak.highlevel.Array(
                cp.array([5, 2, 2, 3, 9, 0, 1], dtype=cp.int32), check_valid=True
            )
        ].form
    )
    assert to_list(
        cuda_array[
            ak.highlevel.Array(cp.array([5, 2, 2, 3, 9, 0, 1]), check_valid=True)
        ]
    ) == [
        5.5,
        2.2,
        2.2,
        3.3,
        9.9,
        0.0,
        1.1,
    ]
    assert (
        cuda_array.to_typetracer()[
            cp.array(ak.highlevel.Array([5, 2, 2, 3, 9, 0, 1]))
        ].form
        == cuda_array[cp.array(ak.highlevel.Array([5, 2, 2, 3, 9, 0, 1]))].form
    )

    array3 = ak.contents.NumpyArray(
        np.array([False, False, False, False, False, True, False, True, False, True])
    )
    cuda_array3 = ak.to_backend(array3, "cuda")

    assert to_list(cuda_array[cuda_array3]) == [5.5, 7.7, 9.9]
    assert cuda_array.to_typetracer()[cuda_array3].form == cuda_array[cuda_array3].form

    content = ak.contents.NumpyArray(np.array([1, 0, 9, 3, 2, 2, 5], dtype=np.int64))
    index = ak.index.Index64(np.array([6, 5, 4, 3, 2, 1, 0], dtype=np.int64))
    indexedarray = ak.contents.IndexedArray(index, content)
    cuda_indexedarray = ak.to_backend(indexedarray, "cuda")

    assert to_list(cuda_array[cuda_indexedarray]) == [5.5, 2.2, 2.2, 3.3, 9.9, 0.0, 1.1]
    assert (
        cuda_array.to_typetracer()[cuda_indexedarray].form
        == cuda_array[cuda_indexedarray].form
    )
    assert to_list(
        cuda_array[cp.array(ak.highlevel.Array(indexedarray, check_valid=True))]
    ) == [
        5.5,
        2.2,
        2.2,
        3.3,
        9.9,
        0.0,
        1.1,
    ]
    assert (
        cuda_array.to_typetracer()[ak.highlevel.Array(cuda_indexedarray)].form
        == cuda_array[ak.highlevel.Array(cuda_indexedarray)].form
    )

    emptyarray = ak.contents.EmptyArray()
    cuda_emptyarray = ak.to_backend(emptyarray, "cuda")

    assert to_list(cuda_array[cuda_emptyarray]) == []
    assert (
        cuda_array.to_typetracer()[cuda_emptyarray].form
        == cuda_array[cuda_emptyarray].form
    )

    array = ak.highlevel.Array(
        np.array([[0.0, 1.1, 2.2, 3.3, 4.4], [5.5, 6.6, 7.7, 8.8, 9.9]]),
        check_valid=True,
    ).layout
    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    numpyarray1 = ak.contents.NumpyArray(np.array([[0, 1], [1, 0]]))
    numpyarray2 = ak.contents.NumpyArray(np.array([[2, 4], [3, 3]]))

    cuda_numpyarray1 = ak.to_backend(numpyarray1, "cuda")
    cuda_numpyarray2 = ak.to_backend(numpyarray2, "cuda")

    assert to_list(
        cuda_array[
            cuda_numpyarray1,
            cuda_numpyarray2,
        ]
    ) == [[2.2, 9.9], [8.8, 3.3]]
    assert (
        cuda_array.to_typetracer()[
            cuda_numpyarray1,
            cuda_numpyarray2,
        ].form
        == cuda_array[
            cuda_numpyarray1,
            cuda_numpyarray2,
        ].form
    )
    assert to_list(cuda_array[cuda_numpyarray1]) == [
        [[0.0, 1.1, 2.2, 3.3, 4.4], [5.5, 6.6, 7.7, 8.8, 9.9]],
        [[5.5, 6.6, 7.7, 8.8, 9.9], [0.0, 1.1, 2.2, 3.3, 4.4]],
    ]
    assert (
        cuda_array.to_typetracer()[cuda_numpyarray1].form
        == cuda_array[cuda_numpyarray1].form
    )


def test_0111_jagged_and_masked_getitem_array_slice_2():
    array = ak.highlevel.Array(
        [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], check_valid=True
    ).layout
    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    content0 = ak.contents.NumpyArray(np.array([5, 2, 2]))
    content1 = ak.contents.NumpyArray(np.array([3, 9, 0, 1]))
    tags = ak.index.Index8(np.array([0, 0, 0, 1, 1, 1, 1], dtype=np.int8))
    index2 = ak.index.Index64(np.array([0, 1, 2, 0, 1, 2, 3], dtype=np.int64))
    unionarray = ak.contents.UnionArray.simplified(tags, index2, [content0, content1])
    cuda_unionarray = ak.to_backend(unionarray, "cuda")

    assert to_list(cuda_array[cuda_unionarray]) == [5.5, 2.2, 2.2, 3.3, 9.9, 0.0, 1.1]
    assert (
        cuda_array.to_typetracer()[cuda_unionarray].form
        == cuda_array[cuda_unionarray].form
    )


def test_0111_jagged_and_masked_getitem_array_slice_with_union():
    array = ak.highlevel.Array(
        [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], check_valid=True
    ).layout
    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    content0 = ak.contents.NumpyArray(np.array([5, 2, 2]))
    content1 = ak.contents.NumpyArray(np.array([3, 9, 0, 1]))
    tags = ak.index.Index8(np.array([0, 0, 0, 1, 1, 1, 1], dtype=np.int8))
    index2 = ak.index.Index64(np.array([0, 1, 2, 0, 1, 2, 3], dtype=np.int64))
    unionarray = ak.contents.UnionArray.simplified(tags, index2, [content0, content1])
    cuda_unionarray = ak.to_backend(unionarray, "cuda")

    assert to_list(
        cuda_array[ak.highlevel.Array(cuda_unionarray, check_valid=True)]
    ) == [
        5.5,
        2.2,
        2.2,
        3.3,
        9.9,
        0.0,
        1.1,
    ]
    assert (
        cuda_array.to_typetracer()[ak.highlevel.Array(cuda_unionarray)].form
        == cuda_array[ak.highlevel.Array(cuda_unionarray)].form
    )


def test_0111_jagged_and_masked_getitem_jagged_mask():
    array = ak.highlevel.Array(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]], check_valid=True
    ).layout
    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    assert to_list(
        cuda_array[[[True, True, True], [], [True, True], [True], [True, True, True]]]
    ) == [[1.1, 2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]
    assert (
        cuda_array.to_typetracer()[
            [[True, True, True], [], [True, True], [True], [True, True, True]]
        ].form
        == cuda_array[
            [[True, True, True], [], [True, True], [True], [True, True, True]]
        ].form
    )
    assert to_list(
        cuda_array[[[False, True, True], [], [True, True], [True], [True, True, True]]]
    ) == [[2.2, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]
    assert (
        cuda_array.to_typetracer()[
            [[False, True, True], [], [True, True], [True], [True, True, True]]
        ].form
        == cuda_array[
            [[False, True, True], [], [True, True], [True], [True, True, True]]
        ].form
    )
    assert to_list(
        cuda_array[[[True, False, True], [], [True, True], [True], [True, True, True]]]
    ) == [[1.1, 3.3], [], [4.4, 5.5], [6.6], [7.7, 8.8, 9.9]]
    assert (
        cuda_array.to_typetracer()[
            [[True, False, True], [], [True, True], [True], [True, True, True]]
        ].form
        == cuda_array[
            [[True, False, True], [], [True, True], [True], [True, True, True]]
        ].form
    )
    assert to_list(
        cuda_array[[[True, True, True], [], [False, True], [True], [True, True, True]]]
    ) == [[1.1, 2.2, 3.3], [], [5.5], [6.6], [7.7, 8.8, 9.9]]
    assert (
        cuda_array.to_typetracer()[
            [[True, True, True], [], [False, True], [True], [True, True, True]]
        ].form
        == cuda_array[
            [[True, True, True], [], [False, True], [True], [True, True, True]]
        ].form
    )
    assert to_list(
        cuda_array[[[True, True, True], [], [False, False], [True], [True, True, True]]]
    ) == [[1.1, 2.2, 3.3], [], [], [6.6], [7.7, 8.8, 9.9]]
    assert (
        cuda_array.to_typetracer()[
            [[True, True, True], [], [False, False], [True], [True, True, True]]
        ].form
        == cuda_array[
            [[True, True, True], [], [False, False], [True], [True, True, True]]
        ].form
    )


def test_0111_jagged_and_masked_getitem_jagged_missing_mask():
    array = ak.highlevel.Array(
        [[1.1, 2.2, 3.3], [], [4.4, 5.5]], check_valid=True
    ).layout
    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    assert to_list(cuda_array[[[True, True, True], [], [True, True]]]) == [
        [1.1, 2.2, 3.3],
        [],
        [4.4, 5.5],
    ]
    assert (
        cuda_array.to_typetracer()[[[True, True, True], [], [True, True]]].form
        == cuda_array[[[True, True, True], [], [True, True]]].form
    )
    assert to_list(cuda_array[[[True, False, True], [], [True, True]]]) == [
        [1.1, 3.3],
        [],
        [4.4, 5.5],
    ]
    assert (
        cuda_array.to_typetracer()[[[True, False, True], [], [True, True]]].form
        == cuda_array[[[True, False, True], [], [True, True]]].form
    )
    assert to_list(cuda_array[[[True, True, False], [], [False, None]]]) == [
        [1.1, 2.2],
        [],
        [None],
    ]
    assert (
        cuda_array.to_typetracer()[[[True, True, False], [], [False, None]]].form
        == cuda_array[[[True, True, False], [], [False, None]]].form
    )
    assert to_list(cuda_array[[[True, True, False], [], [True, None]]]) == [
        [1.1, 2.2],
        [],
        [4.4, None],
    ]
    assert (
        cuda_array.to_typetracer()[[[True, True, False], [], [True, None]]].form
        == cuda_array[[[True, True, False], [], [True, None]]].form
    )

    assert to_list(cuda_array[[[True, None, True], [], [True, True]]]) == [
        [1.1, None, 3.3],
        [],
        [4.4, 5.5],
    ]
    assert (
        cuda_array.to_typetracer()[[[True, None, True], [], [True, True]]].form
        == cuda_array[[[True, None, True], [], [True, True]]].form
    )
    assert to_list(cuda_array[[[True, None, False], [], [True, True]]]) == [
        [1.1, None],
        [],
        [4.4, 5.5],
    ]
    assert (
        cuda_array.to_typetracer()[[[True, None, False], [], [True, True]]].form
        == cuda_array[[[True, None, False], [], [True, True]]].form
    )

    assert to_list(cuda_array[[[False, None, False], [], [True, True]]]) == [
        [None],
        [],
        [4.4, 5.5],
    ]
    assert (
        cuda_array.to_typetracer()[[[False, None, False], [], [True, True]]].form
        == cuda_array[[[False, None, False], [], [True, True]]].form
    )
    assert to_list(cuda_array[[[True, True, False], [], [False, True]]]) == [
        [1.1, 2.2],
        [],
        [5.5],
    ]
    assert (
        cuda_array.to_typetracer()[[[True, True, False], [], [False, True]]].form
        == cuda_array[[[True, True, False], [], [False, True]]].form
    )
    assert to_list(cuda_array[[[True, True, None], [], [False, True]]]) == [
        [1.1, 2.2, None],
        [],
        [5.5],
    ]
    assert (
        cuda_array.to_typetracer()[[[True, True, None], [], [False, True]]].form
        == cuda_array[[[True, True, None], [], [False, True]]].form
    )
    assert to_list(cuda_array[[[True, True, False], [None], [False, True]]]) == [
        [1.1, 2.2],
        [None],
        [5.5],
    ]
    assert (
        cuda_array.to_typetracer()[[[True, True, False], [None], [False, True]]].form
        == cuda_array[[[True, True, False], [None], [False, True]]].form
    )


def test_0111_jagged_and_masked_getitem_masked_of_jagged_of_whatever():
    content = ak.contents.NumpyArray(np.arange(2 * 3 * 5))
    regulararray1 = ak.contents.RegularArray(content, 5, zeros_length=0)
    regulararray2 = ak.contents.RegularArray(regulararray1, 3, zeros_length=0)
    cuda_regulararray2 = ak.to_backend(regulararray2, "cuda", highlevel=False)

    array1 = ak.highlevel.Array(
        [[[2], None, [-1, 2, 0]], [[-3], None, [-5, -3, 4]]], check_valid=True
    ).layout
    array2 = ak.highlevel.Array(
        [[[2], None, [-1, None, 0]], [[-3], None, [-5, None, 4]]],
        check_valid=True,
    ).layout

    cuda_array1 = ak.to_backend(array1, backend="cuda")
    cuda_array2 = ak.to_backend(array2, backend="cuda")

    assert to_list(cuda_regulararray2[cuda_array1]) == [
        [[2], None, [14, 12, 10]],
        [[17], None, [25, 27, 29]],
    ]
    assert (
        cuda_regulararray2.to_typetracer()[cuda_array1].form
        == cuda_regulararray2[cuda_array1].form
    )

    assert to_list(cuda_regulararray2[cuda_array2]) == [
        [[2], None, [14, None, 10]],
        [[17], None, [25, None, 29]],
    ]
    assert (
        cuda_regulararray2.to_typetracer()[cuda_array2].form
        == cuda_regulararray2[cuda_array2].form
    )


def test_0111_jagged_and_masked_getitem_missing():
    array = ak.highlevel.Array(
        [0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9], check_valid=True
    ).layout
    array2 = ak.highlevel.Array([3, 6, None, None, -2, 6], check_valid=True).layout

    cuda_array = ak.to_backend(array, "cuda", highlevel=False)
    cuda_array2 = ak.to_backend(array2, "cuda")

    assert to_list(cuda_array[cuda_array2]) == [
        3.3,
        6.6,
        None,
        None,
        8.8,
        6.6,
    ]
    assert cuda_array.to_typetracer()[cuda_array2].form == cuda_array[cuda_array2].form

    content = ak.contents.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.0, 11.1, 999])
    )
    regulararray = ak.contents.RegularArray(content, 4, zeros_length=0)
    cuda_regulararray = ak.to_backend(regulararray, "cuda", highlevel=False)

    assert to_list(cuda_regulararray) == [
        [0.0, 1.1, 2.2, 3.3],
        [4.4, 5.5, 6.6, 7.7],
        [8.8, 9.9, 10.0, 11.1],
    ]
    array3 = ak.highlevel.Array([2, 1, 1, None, -1], check_valid=True).layout
    cuda_array3 = ak.to_backend(array3, "cuda")

    assert to_list(cuda_regulararray[cuda_array3]) == [
        [8.8, 9.9, 10.0, 11.1],
        [4.4, 5.5, 6.6, 7.7],
        [4.4, 5.5, 6.6, 7.7],
        None,
        [8.8, 9.9, 10.0, 11.1],
    ]
    assert (
        cuda_regulararray.to_typetracer()[cuda_array3].form
        == cuda_regulararray[cuda_array3].form
    )
    assert to_list(cuda_regulararray[:, cuda_array3]) == [
        [2.2, 1.1, 1.1, None, 3.3],
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]
    assert (
        cuda_regulararray.to_typetracer()[:, cuda_array3].form
        == cuda_regulararray[:, cuda_array3].form
    )
    assert to_list(cuda_regulararray[1:, cuda_array3]) == [
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]
    assert (
        cuda_regulararray.to_typetracer()[1:, cuda_array3].form
        == cuda_regulararray[1:, cuda_array3].form
    )

    maskedarray = np.ma.MaskedArray(
        [2, 1, 1, 999, -1], [False, False, False, True, False]
    )
    cuda_maskedarray = ak.to_backend(maskedarray, backend="cuda")

    assert to_list(cuda_regulararray[cuda_maskedarray]) == [
        [8.8, 9.9, 10.0, 11.1],
        [4.4, 5.5, 6.6, 7.7],
        [4.4, 5.5, 6.6, 7.7],
        None,
        [8.8, 9.9, 10.0, 11.1],
    ]
    assert (
        cuda_regulararray.to_typetracer()[cuda_maskedarray].form
        == cuda_regulararray[cuda_maskedarray].form
    )
    assert to_list(
        cuda_regulararray[
            :,
            cuda_maskedarray,
        ]
    ) == [
        [2.2, 1.1, 1.1, None, 3.3],
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]
    assert (
        cuda_regulararray.to_typetracer()[
            :,
            cuda_maskedarray,
        ].form
        == cuda_regulararray[
            :,
            cuda_maskedarray,
        ].form
    )

    assert to_list(
        cuda_regulararray[
            1:,
            cuda_maskedarray,
        ]
    ) == [[6.6, 5.5, 5.5, None, 7.7], [10.0, 9.9, 9.9, None, 11.1]]
    assert (
        cuda_regulararray.to_typetracer()[
            1:,
            cuda_maskedarray,
        ].form
        == cuda_regulararray[
            1:,
            cuda_maskedarray,
        ].form
    )

    content = ak.contents.NumpyArray(
        np.array([[0.0, 1.1, 2.2, 3.3], [4.4, 5.5, 6.6, 7.7], [8.8, 9.9, 10.0, 11.1]])
    )
    cuda_content = ak.to_backend(content, "cuda", highlevel=False)

    assert to_list(cuda_content[cuda_array3]) == [
        [8.8, 9.9, 10.0, 11.1],
        [4.4, 5.5, 6.6, 7.7],
        [4.4, 5.5, 6.6, 7.7],
        None,
        [8.8, 9.9, 10.0, 11.1],
    ]
    assert (
        cuda_content.to_typetracer()[cuda_array3].form == cuda_content[cuda_array3].form
    )
    assert to_list(cuda_content[:, cuda_array3]) == [
        [2.2, 1.1, 1.1, None, 3.3],
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]
    assert (
        cuda_content.to_typetracer()[:, cuda_array3].form
        == cuda_content[:, cuda_array3].form
    )
    assert to_list(cuda_content[1:, cuda_array3]) == [
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]
    assert (
        cuda_content.to_typetracer()[1:, cuda_array3].form
        == cuda_content[1:, cuda_array3].form
    )

    assert to_list(cuda_content[cuda_maskedarray]) == [
        [8.8, 9.9, 10.0, 11.1],
        [4.4, 5.5, 6.6, 7.7],
        [4.4, 5.5, 6.6, 7.7],
        None,
        [8.8, 9.9, 10.0, 11.1],
    ]
    assert (
        cuda_content.to_typetracer()[cuda_maskedarray].form
        == cuda_content[cuda_maskedarray].form
    )
    assert to_list(
        cuda_content[
            :,
            cuda_maskedarray,
        ]
    ) == [
        [2.2, 1.1, 1.1, None, 3.3],
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]
    assert (
        cuda_content.to_typetracer()[
            :,
            cuda_maskedarray,
        ].form
        == cuda_content[
            :,
            cuda_maskedarray,
        ].form
    )
    assert to_list(
        cuda_content[
            1:,
            cuda_maskedarray,
        ]
    ) == [[6.6, 5.5, 5.5, None, 7.7], [10.0, 9.9, 9.9, None, 11.1]]
    assert (
        cuda_content.to_typetracer()[
            1:,
            cuda_maskedarray,
        ].form
        == cuda_content[
            1:,
            cuda_maskedarray,
        ].form
    )

    content = ak.contents.NumpyArray(
        np.array([0.0, 1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8, 9.9, 10.0, 11.1, 999])
    )
    offsets = ak.index.Index64(np.array([0, 4, 8, 12], dtype=np.int64))
    listoffsetarray = ak.contents.ListOffsetArray(offsets, content)
    cuda_listoffsetarray = ak.to_backend(listoffsetarray, "cuda", highlevel=False)

    assert to_list(cuda_listoffsetarray) == [
        [0.0, 1.1, 2.2, 3.3],
        [4.4, 5.5, 6.6, 7.7],
        [8.8, 9.9, 10.0, 11.1],
    ]
    assert to_list(cuda_listoffsetarray[:, cuda_array3]) == [
        [2.2, 1.1, 1.1, None, 3.3],
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]
    assert (
        cuda_listoffsetarray.to_typetracer()[:, cuda_array3].form
        == cuda_listoffsetarray[:, cuda_array3].form
    )
    assert to_list(cuda_listoffsetarray[1:, cuda_array3]) == [
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]
    assert (
        cuda_listoffsetarray.to_typetracer()[1:, cuda_array3].form
        == cuda_listoffsetarray[1:, cuda_array3].form
    )

    assert to_list(
        cuda_listoffsetarray[
            :,
            cuda_maskedarray,
        ]
    ) == [
        [2.2, 1.1, 1.1, None, 3.3],
        [6.6, 5.5, 5.5, None, 7.7],
        [10.0, 9.9, 9.9, None, 11.1],
    ]
    assert (
        cuda_listoffsetarray.to_typetracer()[
            :,
            cuda_maskedarray,
        ].form
        == cuda_listoffsetarray[
            :,
            cuda_maskedarray,
        ].form
    )
    assert to_list(
        cuda_listoffsetarray[
            1:,
            cuda_maskedarray,
        ]
    ) == [[6.6, 5.5, 5.5, None, 7.7], [10.0, 9.9, 9.9, None, 11.1]]
    assert (
        cuda_listoffsetarray.to_typetracer()[
            1:,
            cuda_maskedarray,
        ].form
        == cuda_listoffsetarray[
            1:,
            cuda_maskedarray,
        ].form
    )


def test_0111_jagged_and_masked_getitem_new_slices():
    content = ak.contents.NumpyArray(np.array([1, 0, 9, 3, 2, 2, 5], dtype=np.int64))
    index = ak.index.Index64(np.array([6, 5, -1, 3, 2, -1, 0], dtype=np.int64))
    indexedarray = ak.contents.IndexedOptionArray(index, content)
    cuda_indexedarray = ak.to_backend(indexedarray, backend="cuda")

    assert to_list(cuda_indexedarray) == [5, 2, None, 3, 9, None, 1]

    offsets = ak.index.Index64(np.array([0, 4, 4, 7], dtype=np.int64))
    listoffsetarray = ak.contents.ListOffsetArray(offsets, content)
    cuda_listoffsetarray = ak.to_backend(listoffsetarray, backend="cuda")

    assert to_list(cuda_listoffsetarray) == [[1, 0, 9, 3], [], [2, 2, 5]]

    offsets = ak.index.Index64(np.array([1, 4, 4, 6], dtype=np.int64))
    listoffsetarray = ak.contents.ListOffsetArray(offsets, content)
    cuda_listoffsetarray = ak.to_backend(listoffsetarray, backend="cuda")

    assert to_list(cuda_listoffsetarray) == [[0, 9, 3], [], [2, 2]]

    starts = ak.index.Index64(np.array([1, 99, 5], dtype=np.int64))
    stops = ak.index.Index64(np.array([4, 99, 7], dtype=np.int64))
    listarray = ak.contents.ListArray(starts, stops, content)
    cuda_listarray = ak.to_backend(listarray, backend="cuda")

    assert to_list(cuda_listarray) == [[0, 9, 3], [], [2, 5]]


def test_0111_jagged_and_masked_getitem_record():
    array = ak.highlevel.Array(
        [
            {"x": [0, 1, 2], "y": [0.0, 1.1, 2.2, 3.3]},
            {"x": [3, 4, 5, 6], "y": [4.4, 5.5]},
            {"x": [7, 8], "y": [6.6, 7.7, 8.8, 9.9]},
        ],
        check_valid=True,
    ).layout
    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    array2 = ak.highlevel.Array([[-1, 1], [0, 0, 1], [-1, -2]], check_valid=True).layout
    array3 = ak.highlevel.Array(
        [[-1, 1], [0, 0, None, 1], [-1, -2]], check_valid=True
    ).layout
    array4 = ak.highlevel.Array([[-1, 1], None, [-1, -2]], check_valid=True).layout
    cuda_array2 = ak.to_backend(array2, "cuda")
    cuda_array3 = ak.to_backend(array3, "cuda")
    cuda_array4 = ak.to_backend(array4, "cuda")

    assert to_list(cuda_array[cuda_array2]) == [
        {"x": [2, 1], "y": [3.3, 1.1]},
        {"x": [3, 3, 4], "y": [4.4, 4.4, 5.5]},
        {"x": [8, 7], "y": [9.9, 8.8]},
    ]
    assert cuda_array.to_typetracer()[cuda_array2].form == cuda_array[cuda_array2].form
    assert to_list(cuda_array[cuda_array3]) == [
        {"x": [2, 1], "y": [3.3, 1.1]},
        {"x": [3, 3, None, 4], "y": [4.4, 4.4, None, 5.5]},
        {"x": [8, 7], "y": [9.9, 8.8]},
    ]
    assert cuda_array.to_typetracer()[cuda_array3].form == cuda_array[cuda_array3].form
    assert to_list(cuda_array[cuda_array4]) == [
        {"x": [2, 1], "y": [3.3, 1.1]},
        None,
        {"x": [8, 7], "y": [9.9, 8.8]},
    ]
    assert cuda_array.to_typetracer()[cuda_array4].form == cuda_array[cuda_array4].form


def test_0111_jagged_and_masked_getitem_records_missing():
    array = ak.highlevel.Array(
        [
            {"x": 0, "y": 0.0},
            {"x": 1, "y": 1.1},
            {"x": 2, "y": 2.2},
            {"x": 3, "y": 3.3},
            {"x": 4, "y": 4.4},
            {"x": 5, "y": 5.5},
            {"x": 6, "y": 6.6},
            {"x": 7, "y": 7.7},
            {"x": 8, "y": 8.8},
            {"x": 9, "y": 9.9},
        ],
        check_valid=True,
    ).layout
    array2 = ak.highlevel.Array([3, 1, None, 1, 7], check_valid=True).layout

    cuda_array = ak.to_backend(array, "cuda", highlevel=False)
    cuda_array2 = ak.to_backend(array2, "cuda")

    assert to_list(cuda_array[cuda_array2]) == [
        {"x": 3, "y": 3.3},
        {"x": 1, "y": 1.1},
        None,
        {"x": 1, "y": 1.1},
        {"x": 7, "y": 7.7},
    ]
    assert cuda_array.to_typetracer()[cuda_array2].form == cuda_array[cuda_array2].form

    array = ak.highlevel.Array(
        [
            [
                {"x": 0, "y": 0.0},
                {"x": 1, "y": 1.1},
                {"x": 2, "y": 2.2},
                {"x": 3, "y": 3.3},
            ],
            [
                {"x": 4, "y": 4.4},
                {"x": 5, "y": 5.5},
                {"x": 6, "y": 6.6},
                {"x": 7, "y": 7.7},
                {"x": 8, "y": 8.8},
                {"x": 9, "y": 9.9},
            ],
        ],
        check_valid=True,
    ).layout
    array2 = ak.highlevel.Array([1, None, 2, -1], check_valid=True).layout

    cuda_array = ak.to_backend(array, "cuda", highlevel=False)
    cuda_array2 = ak.to_backend(array2, "cuda")

    assert to_list(cuda_array[:, cuda_array2]) == [
        [{"x": 1, "y": 1.1}, None, {"x": 2, "y": 2.2}, {"x": 3, "y": 3.3}],
        [{"x": 5, "y": 5.5}, None, {"x": 6, "y": 6.6}, {"x": 9, "y": 9.9}],
    ]
    assert (
        cuda_array.to_typetracer()[:, cuda_array2].form
        == cuda_array[:, cuda_array2].form
    )

    array = ak.highlevel.Array(
        [
            {"x": [0, 1, 2, 3], "y": [0.0, 1.1, 2.2, 3.3]},
            {"x": [4, 5, 6, 7], "y": [4.4, 5.5, 6.6, 7.7]},
            {"x": [8, 9, 10, 11], "y": [8.8, 9.9, 10.0, 11.1]},
        ],
        check_valid=True,
    ).layout
    cuda_array = ak.to_backend(array, "cuda", highlevel=False)

    assert to_list(cuda_array[:, cuda_array2]) == [
        {"x": [1, None, 2, 3], "y": [1.1, None, 2.2, 3.3]},
        {"x": [5, None, 6, 7], "y": [5.5, None, 6.6, 7.7]},
        {"x": [9, None, 10, 11], "y": [9.9, None, 10.0, 11.1]},
    ]
    assert (
        cuda_array.to_typetracer()[:, cuda_array2].form
        == cuda_array[:, cuda_array2].form
    )
    assert to_list(cuda_array[1:, cuda_array2]) == [
        {"x": [5, None, 6, 7], "y": [5.5, None, 6.6, 7.7]},
        {"x": [9, None, 10, 11], "y": [9.9, None, 10.0, 11.1]},
    ]
    assert (
        cuda_array.to_typetracer()[1:, cuda_array2].form
        == cuda_array[1:, cuda_array2].form
    )


def test_0111_jagged_and_masked_getitem_regular_regular():
    content = ak.contents.NumpyArray(np.arange(2 * 3 * 5))
    regulararray1 = ak.contents.RegularArray(content, 5, zeros_length=0)
    regulararray2 = ak.contents.RegularArray(regulararray1, 3, zeros_length=0)

    cuda_regulararray2 = ak.to_backend(regulararray2, "cuda", highlevel=False)

    array1 = ak.highlevel.Array(
        [[[2], [1, -2], [-1, 2, 0]], [[-3], [-4, 3], [-5, -3, 4]]],
        check_valid=True,
    ).layout
    array2 = ak.highlevel.Array(
        [[[2], [1, -2], [-1, None, 0]], [[-3], [-4, 3], [-5, None, 4]]],
        check_valid=True,
    ).layout

    cuda_array1 = ak.to_backend(array1, "cuda")
    cuda_array2 = ak.to_backend(array2, "cuda")

    assert to_list(cuda_regulararray2[cuda_array1]) == [
        [[2], [6, 8], [14, 12, 10]],
        [[17], [21, 23], [25, 27, 29]],
    ]
    assert (
        cuda_regulararray2.to_typetracer()[cuda_array1].form
        == cuda_regulararray2[cuda_array1].form
    )

    assert to_list(cuda_regulararray2[cuda_array2]) == [
        [[2], [6, 8], [14, None, 10]],
        [[17], [21, 23], [25, None, 29]],
    ]
    assert (
        cuda_regulararray2.to_typetracer()[cuda_array2].form
        == cuda_regulararray2[cuda_array2].form
    )


def test_0111_jagged_and_masked_getitem_sequential():
    array = ak.highlevel.Array(
        np.arange(2 * 3 * 5).reshape(2, 3, 5).tolist(), check_valid=True
    ).layout
    array2 = ak.highlevel.Array([[2, 1, 0], [2, 1, 0]], check_valid=True).layout

    cuda_array = ak.to_backend(array, "cuda", highlevel=False)
    cuda_array2 = ak.to_backend(array2, "cuda")

    assert to_list(cuda_array[cuda_array2]) == [
        [[10, 11, 12, 13, 14], [5, 6, 7, 8, 9], [0, 1, 2, 3, 4]],
        [[25, 26, 27, 28, 29], [20, 21, 22, 23, 24], [15, 16, 17, 18, 19]],
    ]
    assert cuda_array.to_typetracer()[cuda_array2].form == cuda_array[cuda_array2].form
    assert to_list(cuda_array[cuda_array2, :2]) == [
        [[10, 11], [5, 6], [0, 1]],
        [[25, 26], [20, 21], [15, 16]],
    ]
    assert (
        cuda_array.to_typetracer()[cuda_array2, :2].form
        == cuda_array[cuda_array2, :2].form
    )


def test_0111_jagged_and_masked_getitem_union():
    one = ak.operations.from_iter(
        [["1.1", "2.2", "3.3"], [], ["4.4", "5.5"]], highlevel=False
    )
    two = ak.operations.from_iter(
        [[6.6], [7.7, 8.8], [], [9.9, 10.0, 11.1, 12.2]], highlevel=False
    )
    tags = ak.index.Index8(np.array([0, 0, 0, 1, 1, 1, 1], dtype=np.int8))
    index = ak.index.Index64(np.array([0, 1, 2, 0, 1, 2, 3], dtype=np.int64))
    unionarray = ak.contents.UnionArray(tags, index, [one, two])

    cuda_unionarray = ak.to_backend(unionarray, "cuda")

    assert to_list(cuda_unionarray) == [
        ["1.1", "2.2", "3.3"],
        [],
        ["4.4", "5.5"],
        [6.6],
        [7.7, 8.8],
        [],
        [9.9, 10.0, 11.1, 12.2],
    ]


def test_0111_jagged_and_masked_getitem_union_2():
    one = ak.operations.from_iter([[1.1, 2.2, 3.3], [], [4.4, 5.5]], highlevel=False)
    two = ak.operations.from_iter(
        [[6.6], [7.7, 8.8], [], [9.9, 10.0, 11.1, 12.2]], highlevel=False
    )
    tags = ak.index.Index8(np.array([0, 0, 0, 1, 1, 1, 1], dtype=np.int8))
    index = ak.index.Index64(np.array([0, 1, 2, 0, 1, 2, 3], dtype=np.int64))
    unionarray = ak.contents.UnionArray.simplified(tags, index, [one, two])
    array = ak.highlevel.Array(
        [[0, -1], [], [1, 1], [], [-1], [], [1, -2, -1]], check_valid=True
    ).layout

    cuda_unionarray = ak.to_backend(unionarray, "cuda", highlevel=False)
    cuda_array = ak.to_backend(array, "cuda")

    assert to_list(cuda_unionarray[cuda_array]) == [
        [1.1, 3.3],
        [],
        [5.5, 5.5],
        [],
        [8.8],
        [],
        [10.0, 11.1, 12.2],
    ]
    assert (
        cuda_unionarray.to_typetracer()[cuda_array].form
        == cuda_unionarray[cuda_array].form
    )
